# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import os

import absl.app
import absl.flags
import nbconvert
import nbformat
import ruamel.yaml
import smart_open
from absl import logging
from ruamel.yaml.compat import StringIO as YAMLStringIO

from dyff.api import storage

# -----------------------------------------------------------------------------

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string(
    "audit_yaml", None, "Path to a YAML file containing the Audit manifest."
)
absl.flags.mark_flag_as_required("audit_yaml")

# -----------------------------------------------------------------------------


def main(_unused_argv):
    logging.set_verbosity(logging.INFO)

    yaml = ruamel.yaml.YAML()
    with open(FLAGS.audit_yaml, "r") as fin:
        audit_yaml = yaml.load(fin)
    yaml_string = YAMLStringIO()
    yaml.dump(audit_yaml, yaml_string)
    logging.info(f"audit_yaml:\n{yaml_string.getvalue()}")

    audit_id = audit_yaml["metadata"]["labels"]["audit"]
    auditprocedure_id = audit_yaml["spec"]["auditProcedure"]
    inferenceservice_id = audit_yaml["spec"]["inferenceService"]
    os.environ["ALIGNMENTLABS_AUDIT_SYSTEM_UNDER_TEST"] = inferenceservice_id

    with smart_open.open(
        storage.paths.auditprocedure_notebook(auditprocedure_id), "r"
    ) as fin:
        nb = nbformat.read(fin, as_version=4)

    execute_preprocessor = nbconvert.preprocessors.ExecutePreprocessor()
    execute_body, execute_resources = execute_preprocessor.preprocess(nb)

    html_exporter = nbconvert.HTMLExporter(exclude_input=True)
    html_body, resources = html_exporter.from_notebook_node(execute_body)

    with smart_open.open(storage.paths.auditreport_html(audit_id), "w") as fout:
        fout.writelines(html_body)


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import asyncio
import concurrent.futures
import json
import queue
import random
import re
import threading
import uuid
from collections import Counter, defaultdict
from typing import (
    Any,
    Iterable,
    MutableMapping,
    NamedTuple,
    Optional,
    Sequence,
    TypeVar,
)

import absl.app
import absl.flags
import aiohttp
import pyarrow
import pyarrow.compute
import ruamel.yaml
from absl import logging
from ruamel.yaml.compat import StringIO as YAMLStringIO
from typing_extensions import TypeAlias

from dyff.api import storage
from dyff.schema import ids
from dyff.schema.adapters import Adapter, HTTPData, Pipeline, create_pipeline
from dyff.schema.dataset import arrow

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_bool(
    "dryrun",
    False,
    "Scan the dataset but don't connect or make requests to the model service.",
)

absl.flags.DEFINE_string(
    "evaluation_yaml", None, "Path to a YAML file containing the Evaluation manifest."
)

absl.flags.DEFINE_bool("fail_randomly", False, "Insert random failures for testing.")


ERRORCODE_SUCCESS = 0
ERRORCODE_OUTPUT = 10
ERRORCODE_INCOMPLETE = 20


DataItem: TypeAlias = dict[str, Any]


class ItemID(NamedTuple):
    replication: str
    item_index: int


def _start_background_loop(loop: asyncio.AbstractEventLoop) -> None:
    """Entrypoint for the background thread that runs asyncio."""
    asyncio.set_event_loop(loop)
    loop.run_forever()


async def _rpc_inference(
    session: aiohttp.ClientSession,
    url: str,
    item_id: ItemID,
    http_data: HTTPData,
) -> tuple[ItemID, HTTPData]:
    """Post an inference request.

    Parameters:
      session: Shared ClientSession instance
      url: URL of the inference service
      item_id: The unique identifier of the instance
      http_payload: Tuple containing 'headers' and 'data' fields for the request

    Returns:
      (item_id, response_json): ``item_id`` is the same as the input item_id,
        ``response_json`` contains the response body as a JSON object.
    """
    timeout = aiohttp.ClientTimeout(total=30)
    headers = {"content-type": http_data.content_type}
    async with session.post(
        url,
        headers=headers,
        json=http_data.data,
        timeout=timeout,
        raise_for_status=True,
    ) as response:
        response_json = await response.json()
        response_data = HTTPData("application/json", response_json)
        return (item_id, response_data)


async def _throttled_rpc_inference(semaphore: asyncio.Semaphore, *args, **kwargs):
    async with semaphore:
        return await _rpc_inference(*args, **kwargs)


class _Monitor:
    def __init__(self):
        self._counts = defaultdict(int)
        self._totals = defaultdict(int)

    def event(self, name: str):
        self._counts[name] += 1
        self._totals[name] += 1

    @property
    def totals(self):
        return dict(self._totals)

    def totals_for_event(self, event: str):
        return self._totals.get(event, 0)

    async def __call__(self):
        interval_seconds = 60.0
        while True:
            self._counts.clear()
            await asyncio.sleep(interval_seconds)
            for k, v in sorted(self._counts.items()):
                logging.info(f"Monitor: {k}: {v/interval_seconds}/sec")


class RetryManager:
    def __init__(self):
        self._items: dict[ItemID, DataItem] = {}
        self._counts: MutableMapping[ItemID, int] = defaultdict(int)
        self._failures: list[ItemID] = []

    def retry(self, item_id: ItemID, item) -> bool:
        if self._items.get(item_id) is not None:
            logging.error(f"item {item_id} already queued")
            raise ValueError(f"item {item_id} already queued")
        self._counts[item_id] += 1
        if self._counts[item_id] >= 100:
            logging.error(f"item {item_id} retried 100 times; apparently unprocessable")
            self._failures.append(item_id)
            return False
        else:
            if self._counts[item_id] == 10:
                logging.warning(f"retrying {item_id} for the 10th time")
            self._items[item_id] = item
            return True

    def get(self) -> tuple[ItemID | None, DataItem | None]:
        if len(self._items) == 0:
            return None, None
        item_id = random.choice(list(self._items.keys()))
        item = self._items[item_id]
        del self._items[item_id]
        return item_id, item

    def failures(self):
        return self._failures


def _start_workers(
    *,
    num_workers: int,
    monitor: _Monitor,
    session: aiohttp.ClientSession,
    url: str,
    input_adapter: Adapter,
    output_adapter: Adapter,
    in_queue: asyncio.Queue[tuple[ItemID, DataItem]],
    out_queue: asyncio.Queue[tuple[ItemID, list[DataItem] | None]],
    retry_manager: RetryManager,
):
    """Start the asyncio inference workers."""
    tasks = []
    for i in range(num_workers):
        tasks.append(
            asyncio.create_task(
                _worker(
                    worker_id=i,
                    monitor=monitor,
                    session=session,
                    url=url,
                    input_adapter=input_adapter,
                    output_adapter=output_adapter,
                    in_queue=in_queue,
                    out_queue=out_queue,
                    retry_manager=retry_manager,
                )
            )
        )
    return tasks


async def _worker(
    *,
    worker_id: int,
    monitor: _Monitor,
    session: aiohttp.ClientSession,
    url: str,
    input_adapter: Adapter,
    output_adapter: Adapter,
    in_queue: asyncio.Queue[tuple[ItemID, DataItem]],
    out_queue: asyncio.Queue[tuple[ItemID, list[DataItem] | None]],
    retry_manager: RetryManager,
):
    """Pulls instances from the input queue and places inference results on
    the output queue. Processing is complete once ``in_queue.join()`` returns.
    """
    while True:
        item_id, item = retry_manager.get()
        if item_id is not None:
            monitor.event("retry")
        else:
            item_id, item = await in_queue.get()
            monitor.event("input")
        assert item is not None

        processed = False
        failure = False
        output_items = None
        try:
            input_data = list(input_adapter([item]))
            if len(input_data) != 1:
                raise AssertionError(
                    f"input pipeline yielded multiple values for {item_id}"
                )
            input_item = input_data[0]
            # TODO: What if the input isn't JSON?
            response_id, response_data = await _rpc_inference(
                session, url, item_id, HTTPData("application/json", input_item)
            )
            if item_id != response_id:
                raise AssertionError(
                    f"wires crossed: input {item_id} != output {response_id}"
                )
            # This throws a subclass of dyff.models.outputs.TaskOutputError
            # on verification failure
            output_items = list(output_adapter([response_data.data]))
            processed = True
        except aiohttp.ClientError as ex:
            logging.debug(f"worker {worker_id}: ClientError; retrying: {item_id}")
            monitor.event(type(ex).__name__)
            # Retry
            failure = not retry_manager.retry(item_id, item)
            # Prevent spamming failed connections if the InferenceServices are
            # still starting
            await asyncio.sleep(30)
        except asyncio.exceptions.TimeoutError:
            logging.debug(f"worker {worker_id}: TimeoutError; retrying: {item_id}")
            monitor.event("TimeoutError")
            # Retry
            failure = not retry_manager.retry(item_id, item)
        except Exception as ex:
            event_name = type(ex).__name__
            logging.exception(
                f"worker {worker_id}: {event_name} (fatal): {item_id}: |{response_data}|"
            )
            failure = True
        finally:
            if failure:
                monitor.event("Failure")
                # Failure means there's something unrecoverably wrong with the
                # input instance. (item_id, None) indicates failed inference.
                await out_queue.put((item_id, None))
                in_queue.task_done()
            elif processed:
                await out_queue.put((item_id, output_items))
                in_queue.task_done()
                monitor.event("output")


async def _copy_output(async_queue: asyncio.Queue, sync_queue: queue.Queue):
    """Copies output from the asyncio queue to the synchronous queue. All of
    the output has been copied once ``await async_queue.join()`` returns.
    """
    pending_output = None
    while True:
        if pending_output is None:
            pending_output = await async_queue.get()
            assert pending_output is not None

        try:
            sync_queue.put_nowait(pending_output)
        except queue.Full:
            await asyncio.sleep(0.1)
        else:
            pending_output = None
            async_queue.task_done()


async def _async_main(
    *,
    num_workers: int,
    url: str,
    input_adapter: Adapter,
    output_adapter: Adapter,
    in_queue: queue.Queue[tuple[ItemID, DataItem] | None],
    out_queue: queue.Queue[tuple[ItemID | None, list[DataItem] | None]],
):
    """The main function of the asyncio thread. Spawns workers and output
    task and waits for them to complete.

    Parameters:
      num_workers: Number of worker tasks
      url: URL of the inference service
      in_queue: Instances for inference. Input is complete when
        ``in_queue.get()`` returns the sentinel value ``None``.
      out_queue: Inference results. Inference is complete when
        ``out_queue.get()`` returns the sentinel value ``(None, None)``.
    """
    # Kubernetes does not load-balance persistent connections properly;
    # force_close=True makes them not persistent.
    # See: https://stackoverflow.com/a/71216872
    async with aiohttp.ClientSession(
        connector=aiohttp.TCPConnector(force_close=True)
    ) as session:
        # Propagate input queue limits
        work_queue: asyncio.Queue[tuple[ItemID, DataItem]] = asyncio.Queue(
            in_queue.maxsize
        )
        result_queue: asyncio.Queue[tuple[ItemID, list[DataItem] | None]] = (
            asyncio.Queue()
        )
        retry_manager = RetryManager()
        monitor = _Monitor()
        processed_indices: set[ItemID] = set()

        worker_tasks = []
        worker_tasks.append(asyncio.create_task(_copy_output(result_queue, out_queue)))
        worker_tasks.append(asyncio.create_task(monitor()))
        worker_tasks.extend(
            _start_workers(
                num_workers=num_workers,
                monitor=monitor,
                session=session,
                url=url,
                input_adapter=input_adapter,
                output_adapter=output_adapter,
                in_queue=work_queue,
                out_queue=result_queue,
                retry_manager=retry_manager,
            )
        )

        while True:
            try:
                work_item = in_queue.get_nowait()
                if work_item is None:  # sentinel
                    break
                item_id, item = work_item
                if item_id in processed_indices:
                    logging.error(f"duplicate item {item_id} in in_queue")
                    monitor.event("DuplicateWorkItem")
                else:
                    processed_indices.add(item_id)
                    await work_queue.put((item_id, item))
            except queue.Empty:
                await asyncio.sleep(0.1)

        # Wait for output
        logging.info("joining work_queue")
        await work_queue.join()
        logging.info("joining result_queue")
        await result_queue.join()
        # Cancel tasks
        logging.info("cancelling worker_tasks")
        for task in worker_tasks:
            task.cancel()
        logging.info("waiting for canceled tasks")
        await asyncio.gather(*worker_tasks, return_exceptions=True)
        logging.info("putting sentinel on out_queue")
        out_queue.put((None, None))  # sentinel

        logging.info(f"Monitor event totals: {monitor.totals}")
        if monitor.totals_for_event("input") != monitor.totals_for_event("output"):
            logging.error("Monitor: 'input' != 'output'")
        failures = retry_manager.failures()
        if len(failures) > 0:
            logging.error(f"failed indices: {failures}")


def _output_thread(
    *,
    out_queue: queue.Queue[tuple[ItemID | None, list[DataItem] | None]],
    output_path: str,
    output_schema: pyarrow.Schema,
    batch_size: int = 1000,
):
    """Entry point for the thread that writes PyArrow output. Assumes that
    inference is finished if ``out_queue.get()`` returns the sentinel
    ``(None, None)``.
    """

    def batch_generator(output_schema):
        processed = set()
        batch_idx = 0
        result_batch = []
        while True:
            item_id, responses = out_queue.get()
            if item_id is None:  # sentinel
                logging.info(f"output {len(processed)} unique indices")
                break
            elif responses is None:
                logging.error(f"inference failed: item {item_id}")
                continue

            if item_id in processed:
                logging.error(f"duplicate item {item_id} in out_queue")
                continue
            else:
                processed.add(item_id)

            logging.debug(f"finished {item_id}: {len(responses)} responses")
            for i, response in enumerate(responses):
                response["_response_index_"] = i
            response_record = {
                "_replication_": item_id.replication,
                "_index_": item_id.item_index,
                "responses": responses,
            }
            result_batch.append(response_record)

            if len(result_batch) >= batch_size:
                logging.info(f"finished batch {batch_idx}")
                batch_idx += 1
                yield pyarrow.RecordBatch.from_pylist(
                    result_batch, schema=output_schema
                )
                result_batch = []
        if len(result_batch) > 0:
            logging.info(f"finished batch {batch_idx} (final batch)")
            yield pyarrow.RecordBatch.from_pylist(result_batch, schema=output_schema)

    # Note: We don't partition the outputs because that undermines our goal of
    # writing small chunks frequently to preserve progress. We can re-partition
    # in verify_evaluation_output if desired.
    arrow.write_dataset(
        batch_generator(output_schema),
        output_path=output_path,
        feature_schema=output_schema,
        # Force frequent writes so we don't lose progress
        max_rows_per_file=batch_size,
        # pyarrow.lib.ArrowInvalid: max_rows_per_group must be less than or equal to max_rows_per_file
        max_rows_per_group=batch_size,
        # Use UUIDs so that we never overwrite old data. On a restart, we read
        # all the data that's already persisted and skip those indices.
        basename_template=f"{uuid.uuid4()}.{{i}}.parquet",
    )


class _MissingItemPredicate:
    def __init__(self, *, not_missing: Optional[set[ItemID]] = None):
        self._not_missing = not_missing or set()

    def __call__(self, item_id: ItemID) -> bool:
        return item_id not in self._not_missing


def _dataset_fragments(path: str) -> list[str]:
    """Returns a list of full GCS paths to each dataset fragment for which
    ``path`` is a prefix. Remember that GCS "directory" list operations are
    always recursive because they're not really "directories".
    """
    # FIXME: Should open the .dyff/artifacts.json file to get the artifact list,
    # but this is not created for evaluation output datasets currently.
    objects = storage.list_dir(path)
    pattern = r"/[^/]*\.parquet$"
    fragments = [obj for obj in objects if re.search(pattern, obj)]
    return fragments


def _missing_item_predicate(output_path: str):
    try:
        # There is a race between output_path creation and reading the
        # dataset here. If the directory has been created, we open an empty
        # dataset and PyArrow throws an exception because there is no _index_
        # column. Thus, we need to check for an empty dataset.
        fragments = _dataset_fragments(output_path)
        if len(fragments) == 0:
            return _MissingItemPredicate()
        outputs = arrow.open_dataset(output_path)
        not_missing: list[ItemID] = []
        for b in outputs.to_batches(columns=["_replication_", "_index_"]):
            not_missing.extend(
                ItemID(replication=x["_replication_"], item_index=x["_index_"])
                for x in b.to_pylist()
            )
        logging.info(f"found {len(not_missing)} outputs in {output_path}")
        counts = Counter(not_missing)
        duplicates = [k for k, v in counts.items() if v > 1]
        if len(duplicates) > 0:
            logging.error(f"duplicates in outputs: {duplicates}")
            raise ValueError("duplicates in outputs")
        return _MissingItemPredicate(not_missing=set(not_missing))
    except FileNotFoundError:
        return _MissingItemPredicate()


def _inference_service_address(evaluation_yaml) -> str:
    name = evaluation_yaml["metadata"]["name"]
    namespace = evaluation_yaml["metadata"]["namespace"]
    return f"http://{name}-infer.{namespace}:80"


def _load_pipeline(pipeline_spec: list[dict]) -> Pipeline:
    pipeline_spec = [
        {
            "kind": adapter_spec["kind"],
            "configuration": json.loads(adapter_spec["configuration"]),
        }
        for adapter_spec in pipeline_spec
    ]
    return create_pipeline(pipeline_spec)


def main(_unused_argv: Sequence[str]) -> Optional[int]:
    logging.set_verbosity(logging.DEBUG)

    yaml = ruamel.yaml.YAML()
    with open(FLAGS.evaluation_yaml, "r") as fin:
        evaluation = yaml.load(fin)
    yaml_string = YAMLStringIO()
    yaml.dump(evaluation, yaml_string)
    logging.info(f"evaluation_yaml:\n{yaml_string.getvalue()}")

    dataset = evaluation["spec"]["dataset"]
    # TODO: We allow specifying a different path for dataset uploads, so we
    # should allow loading them from an alternative path here. Actually,
    # should we expand Dataset to include its storage location?
    dataset_path = storage.paths.dataset_root(dataset)
    ds = arrow.open_dataset(dataset_path)
    logging.info(f"input dataset: {dataset}")

    interface = evaluation["spec"]["interface"]
    inference_endpoint = interface["endpoint"]
    output_schema = arrow.decode_schema(interface["outputSchema"]["arrowSchema"])
    input_pipeline = _load_pipeline(interface["inputPipeline"])
    output_pipeline = _load_pipeline(interface["outputPipeline"])

    # defaults
    batch_size = 1000
    filter_expression = None
    # if (dataset_config := evaluation["spec"].get("datasetConfiguration")) is not None:
    #     if (filters := dataset_config.get("filters")) is not None:
    #         filter_expression = parse_dataset_filters(filters)
    #         logging.info(f"dataset filter: {filter_expression}")
    #     batch_size = dataset_config.get("batchSize", batch_size)

    input_rows = ds.scanner(filter=filter_expression).count_rows()
    logging.info(f"input dataset: {input_rows} rows")

    evaluation_id = evaluation["spec"]["id"]
    replications: int = evaluation["spec"]["replications"]
    replication_ids = [
        ids.replication_id(evaluation_id, i) for i in range(replications)
    ]

    output_path = storage.paths.outputs_raw(evaluation_id)
    missing = _missing_item_predicate(output_path)
    processed_items = set()

    # defaults
    # FIXME: These should be defaults in the CRD, not here
    inference_service_replicas = evaluation["spec"]["inferenceSession"].get(
        "replicas", 1
    )
    workers_per_replica = evaluation["spec"].get("workersPerReplica", 2)
    num_workers = inference_service_replicas * workers_per_replica

    service_address = _inference_service_address(evaluation)
    service_url = f"{service_address}/{inference_endpoint}"

    # Limit input queue to avoid memory limits
    work_queue: queue.Queue[tuple[ItemID, DataItem] | None] = queue.Queue(
        maxsize=(4 * batch_size)
    )
    result_queue: queue.Queue[tuple[ItemID | None, list[DataItem] | None]] = (
        queue.Queue()
    )

    logging.info("starting asyncio thread")
    asyncio_loop = asyncio.new_event_loop()
    async_thread = threading.Thread(
        target=_start_background_loop, args=(asyncio_loop,), daemon=True
    )
    async_thread.start()
    logging.info("launching _async_main")
    async_main = asyncio.run_coroutine_threadsafe(
        _async_main(
            num_workers=num_workers,
            url=service_url,
            input_adapter=input_pipeline,
            output_adapter=output_pipeline,
            in_queue=work_queue,
            out_queue=result_queue,
        ),
        asyncio_loop,
    )

    output_error = False
    with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
        # Doing this with concurrent.futures is the simplest way to propagate
        # exceptions from the thread.
        # See: https://docs.python.org/3/library/concurrent.futures.html#threadpoolexecutor-example
        logging.info("starting output thread")
        output_future = executor.submit(
            _output_thread,
            out_queue=result_queue,
            output_path=output_path,
            output_schema=output_schema,
            batch_size=batch_size,
        )

        T = TypeVar("T")

        def shuffle(items: Iterable[T], buffer_size: int = 10000) -> Iterable[T]:
            buffer: list[T] = []
            for item in items:
                if len(buffer) == buffer_size:
                    idx = random.randrange(buffer_size)
                    yield buffer[idx]
                    buffer[idx] = item
                else:
                    buffer.append(item)
            # Yield remaining items
            random.shuffle(buffer)
            yield from buffer

        def work_items() -> Iterable[tuple[ItemID, DataItem]]:
            for record_batch in ds.to_batches(
                batch_size=batch_size, filter=filter_expression
            ):
                for record in record_batch.to_pylist():
                    index = record["_index_"]
                    for replication_id in replication_ids:
                        item_id = ItemID(replication=replication_id, item_index=index)
                        if missing(item_id):
                            logging.debug(f"enqueue {item_id}")
                            yield (item_id, record)
                        processed_items.add(item_id)

        logging.info("processing dataset")
        for work_item in shuffle(work_items()):
            work_queue.put(work_item)
        work_queue.put(None)  # sentinel

        logging.info("waiting for output_thread")
        for future in concurrent.futures.as_completed([output_future]):
            try:
                future.result()
            except Exception:
                logging.exception("output error")
                output_error = True
        # Thread gets joined on context exit
        logging.info("joining output_thread")

    logging.info("waiting for async_main")
    async_main.result()
    logging.info("stopping asyncio_loop")
    asyncio_loop.stop()
    # This join never returns, which makes me nervous because I think that it
    # ought to return. Removing it doesn't seem to cause problems, though.
    # logging.info("joining async_thread")
    # async_thread.join()

    if output_error:
        return ERRORCODE_OUTPUT
    else:
        N = input_rows * replications
        logging.info("verifying output")
        output_rows = arrow.open_dataset(output_path).count_rows()

        if output_rows != N:
            logging.error(
                f"incomplete output: output_rows {output_rows} != input_rows*replications {N}"
            )
            return ERRORCODE_INCOMPLETE
        elif len(processed_items) != N:
            logging.error(
                f"incomplete output: processed_items {len(processed_items)}"
                f" != input_rows*replications {N}"
            )
            return ERRORCODE_INCOMPLETE
        else:
            logging.info("Alles gut!")
            return ERRORCODE_SUCCESS


if __name__ == "__main__":
    absl.app.run(main)



#!/usr/bin/env python3
# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import json
import sys

# If we specify documentation with pydantic.Field(description=...) for a
# non-primitive attribute, then fastapi will put the
# ``"$ref": "#/components/schema/..."`` bit inside a superfluous ``"allOf"``.
#
# For a nested enum:
#
#   "sourceKind": {
#     "description": "The kind of source from which the InferenceService is obtained.",
#     "allOf": [
#       {
#         "$ref": "#/components/schemas/InferenceServiceSources"
#       },
#     ]
#
# Or, for a nested object:
#
#   "modelConfiguration": {
#     "title": "Modelconfiguration",
#     "description": "How to configure the Model for the Evaluation (e.g., hyperparameters).",
#     "allOf": [
#       {
#         "$ref": "#/components/schemas/ModelConfiguration"
#       }
#     ]
#   },
#
# This will then break autorest, with a message like:
#
#   warning | PreCheck/SchemaMissingType | The schema 'Evaluation-modelConfiguration' with an undefined type and 'allOf'/'anyOf'/'oneOf' is a bit ambiguous. This has been auto-corrected to 'type:object'
#
# Followed by an error like this:
#
#   error   | PreCheck/AllOfTypeDifferent | Schema 'InferenceService-sourceKind' has an allOf reference to 'InferenceServiceSources' but those schema have different types:
#     - InferenceService-sourceKind: object
#     - InferenceServiceSources: string
#
# This script removes the problematic ``"allOf"``s
#
# It is very fragile and may not work if the structure of the models changes
# significantly.
#
# TODO: File a bug report with pydantic and/or fastapi
# The problematic function is pydantic.schema.get_schema_ref(), which adds
# the "allOf" wrapper because the ``schema_overrides`` flag is ``True``. This,
# in turn, is because pydantic.schema.get_field_info_schema() sets that
# flag to ``True`` because we manually specified ``description``.


def main(argv):
    file_path = argv[1]
    with open(file_path, "r") as fin:
        openapi = json.load(fin)

    # Remove the 'allOf' wrappers
    for schema_name, schema_spec in openapi["components"]["schemas"].items():
        if "properties" in schema_spec:
            for name, spec in schema_spec["properties"].items():
                if "allOf" in spec:
                    ref = spec["allOf"][0]["$ref"]
                    del spec["allOf"]
                    spec["$ref"] = ref

    # Remove patternProperties, which is part of JSON Schema but not supported
    # in OpenAPI 3.0
    entities_with_pattern_properties = [
        "Audit",
        "Dataset",
        "Evaluation",
        "InferenceService",
        "InferenceSession",
        "LabelUpdateRequest",
        "Model",
        "Report",
    ]
    for schema_name, schema_spec in openapi["components"]["schemas"].items():
        if schema_name in entities_with_pattern_properties:
            try:
                labels = schema_spec["properties"]["labels"]
                del labels["patternProperties"]
            except:
                pass

            try:
                annotations = schema_spec["properties"]["annotations"]
                del annotations["patternProperties"]
            except:
                pass

    with open(file_path, "w") as fout:
        json.dump(openapi, fout, indent=2)


if __name__ == "__main__":
    main(sys.argv)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import pyarrow


def Image():
    return pyarrow.struct(
        [
            pyarrow.field("bytes", pyarrow.binary()),
            pyarrow.field("format", pyarrow.string()),
        ]
    )


def TextSpan(name: str):
    return pyarrow.field(
        name,
        pyarrow.struct(
            [
                pyarrow.field(
                    "start",
                    pyarrow.int64(),
                    metadata={
                        "__doc__": "The tag applies to the slice start:end of the text."
                    },
                ),
                pyarrow.field(
                    "end",
                    pyarrow.int64(),
                    metadata={
                        "__doc__": "The tag applies to the slice start:end of the text."
                    },
                ),
                pyarrow.field(
                    "tag",
                    pyarrow.string(),
                    metadata={"__doc__": "The tag of the span."},
                ),
            ]
        ),
        metadata={"__doc__": "Associates a tag with a text slice."},
    )



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import secrets
import warnings
from datetime import datetime, timedelta
from typing import Any, Mapping, Optional, Protocol

import jose.jwt
from passlib.hash import bcrypt

from dyff.schema import ids
from dyff.schema.platform import AccessGrant, APIKey, Entities

from .backend.base.auth import AuthBackend

ACCOUNT_TOKEN = "account"
EPHEMERAL_TOKEN = "ephemeral"

TOKEN_CHOICES = (
    ACCOUNT_TOKEN,
    EPHEMERAL_TOKEN,
)


class AuthenticationError(RuntimeError):
    pass


class Signer(Protocol):
    """Manages signatures for access tokens."""

    def sign_api_key(self, api_key: APIKey) -> str:
        """Generate a signed token for the given APIKey."""

    def unsign_api_key(self, token: str) -> APIKey:
        """Extract an APIKey from a signed token. Raises an error if the
        signature is invalid.
        """


class NoSigner:
    """Raises AuthenticationError on signature operations."""

    def sign_api_key(self, api_key: APIKey) -> str:
        """Generate a signed token for the given APIKey."""
        raise AuthenticationError("no signer")

    def unsign_api_key(self, token: str) -> APIKey:
        """Extract an APIKey from a signed token. Raises an error if the
        signature is invalid.
        """
        raise AuthenticationError("no signer")


class JWTSigner:
    """Encodes and decodes JSON Web Tokens."""

    algorithm = "HS256"

    def __init__(self, signing_secret: str | bytes | None):
        self._secret = signing_secret

    def _headers(self) -> dict:
        return {"typ": "JWT", "alg": JWTSigner.algorithm}

    def sign_api_key(self, api_key: APIKey) -> str:
        """Generate a signed token for the given APIKey."""
        headers = self._headers()
        payload = {
            "jti": api_key.id,
            "sub": api_key.subject,
            "nbf": api_key.created,
            "exp": api_key.expires,
            "iat": datetime.utcnow(),
            "grants": [g.dict() for g in api_key.grants],
        }
        if api_key.secret is not None:
            payload["secret"] = api_key.secret
        return jose.jwt.encode(
            payload, key=self._secret, algorithm=JWTSigner.algorithm, headers=headers
        )

    def unsign_api_key(self, token: str) -> APIKey:
        """Extract an APIKey from a signed token. Raises an error if the
        signature is invalid.
        """
        return self._unsign_api_key(token)

    def _unsign_api_key(
        self, token: str, *, decode_options: Optional[Mapping[str, Any]] = None
    ) -> APIKey:
        """Extract an APIKey from a signed token. Raises an error if the
        signature is invalid.
        """
        options = {"require": ["jti", "sub", "nbf", "exp", "iat"]}
        if decode_options:
            options.update(decode_options)
        payload = jose.jwt.decode(
            token, key=self._secret, algorithms=JWTSigner.algorithm, options=options
        )
        return APIKey.parse_obj(
            dict(
                id=payload["jti"],
                subject=payload["sub"],
                created=payload["nbf"],
                expires=payload["exp"],
                secret=payload.get("secret"),
                grants=payload["grants"],
            )
        )


def get_signer(signing_secret: str | None) -> Signer:
    if signing_secret is not None:
        return JWTSigner(signing_secret)
    else:
        return NoSigner()


def generate_api_key(
    *,
    subject_type: Entities,
    subject_id: str,
    grants: Optional[list[AccessGrant]] = None,
    expires: Optional[datetime] = None,
    generate_secret: bool = False,
) -> APIKey:
    """Generates a new API key for a user. User keys have a secret that uniquely
    identifies the user account.
    """
    if grants is None:
        warnings.warn("Creating an API key with no access grants")
        grants = []
    instance_secret = secrets.token_urlsafe(16) if generate_secret else None
    now = datetime.utcnow()
    if expires is None:
        expires = now + timedelta(days=365)
    return APIKey(
        id=ids.generate_entity_id(),
        subject=f"{subject_type}/{subject_id}",
        created=now,
        expires=expires,
        secret=instance_secret,
        grants=grants,
    )


def hashed_api_key(api_key: APIKey) -> APIKey:
    """Returns a new APIKey where the ``.secret`` field is cryptographically
    hashed.
    """
    return api_key.copy(update={"secret": bcrypt.hash(api_key.secret)})


def verify_api_key_hash(api_key: APIKey, hashed_key: APIKey) -> bool:
    """Returns True if ``api_key.secret`` matches ``hashed_key.secret``."""
    return bcrypt.verify(api_key.secret, hashed_key.secret)


def verify_api_token(
    token: str, signer: Signer, *, auth_backend: Optional[AuthBackend] = None
) -> APIKey:
    """Verifies the token and returns its "payload" as a JSON dict. Raises an
    exception if the signature is invalid or if the secret doesn't match the
    hashed secret in the datastore.
    """
    try:
        unsigned_key = signer.unsign_api_key(token)
    except Exception as ex:
        raise AuthenticationError("bad signature") from ex

    # Only user keys have a secret
    if unsigned_key.secret is not None:
        if auth_backend is None:
            raise AssertionError("no auth_backend provided")
        kind, account, *rest = unsigned_key.subject.split("/")
        if kind != Entities.Account or len(rest) != 0:
            raise AuthenticationError("expected subject = 'Account/id'")
        stored_key = auth_backend.get_api_key(account, unsigned_key.id)
        if (stored_key is None) or (not verify_api_key_hash(unsigned_key, stored_key)):
            raise AuthenticationError("unverified")

    return unsigned_key



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import shutil
import urllib
from pathlib import Path

import requests

from ...api.typing import PathLike


def download_file(url: str, *, local_filename: PathLike = None) -> Path:
    """Download a file from a URL.

    Args:
      url: The URL to download from
      local_filename (optional): Local path to save the downloaded file.
        Default: last element of the URL

    Returns:
      The local filename.

    See:
      https://stackoverflow.com/a/39217788
    """
    parsed_url = urllib.parse.urlparse(url)  # Primarily for validity-checking
    if local_filename is None:
        local_filename = Path(parsed_url.path.split("/")[-1])
    else:
        local_filename = Path(local_filename)
    with requests.get(url, stream=True) as r:
        with open(local_filename, "wb") as f:
            shutil.copyfileobj(r.raw, f)

    return local_filename


def fetch_url(
    url: str, *, local_directory: PathLike, local_file: PathLike = None
) -> Path:
    parsed_url = urllib.parse.urlparse(url)  # Primarily for validity-checking
    remote_file = parsed_url.path.split("/")[-1]
    local_file = local_file or remote_file
    local_path = Path(local_directory) / local_file
    download_file(url, local_filename=local_path)
    return local_path



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

"""Tools for ingesting raw data into Apache Arrow datasets.

We model the raw data as an ``Archive``, which is a generator of
``ArchiveEntry``s. Each ``ArchiveEntry`` consists of a path to a leaf file
in the dataset, and the raw bytes of that file. A ``FileProcessor`` wraps
an ``Archive`` to create a generator of individual instance, represented
as nested dictionaries. This generator then goes to the ``write_dataset``
function of ``pyarrow``.

Processing happens in streaming mode, to accommodate raw data that will not
fit in memory. Processors must not assume that the files arrive in any
particular order, but they should assume that a depth-first order is
the typical case.
"""

from __future__ import annotations

import re
from collections import defaultdict
from pathlib import Path
from typing import Callable, Dict, Generator, List, Optional, Sequence, Union

import numpy as np
import pyarrow
import pyarrow.dataset
import pyarrow.fs
import yaml
from absl import logging
from smart_open import open

from ...core import dynamic_import

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from ...api import storage
from ...api.data import archives
from ...api.typing import PathLike

DataTypeConstructor = Callable[[], pyarrow.DataType]


class PathFeatureExtractor:
    """Extracts individual data instances whose features are specified through
    a combination of path elements and entries in data files.

    Our abstract model of a dataset is a list of one or more "archives". Each
    archive contains a tree of "paths" terminating in "leaf" files. The paths
    encode a subset of the features, such as different "splits"
    of the data or different "intensities" of the same data manipulation. The
    rest of the features are encoded in the leaf files. For example, the leafs
    might be .csv files with columns for different features, or they might be
    .jpeg files that conceptually encode a single "feature" called "image".

    Feature extraction from leafs is delegated to "leaf processors". The
    ``PathFeatureExtractor`` wraps these leaf processors to create a generator
    that extracts features from both leaf files and path elements.

    Once ``leaf()`` has been called, the PathFeatureExtractor is "finished" and
    no more features can be added. After calling ``leaf()``, each call to
    ``__call__(archive)`` returns a generator that yields instances from that
    archive. The extractor adds a special feature named ``_index_`` that
    uniquely identifies each item in the dataset. Note that the archives are
    considered parts of the same dataset, so the index **does not reset**
    between archives.
    """

    def __init__(self):
        self._partitions = []
        self._features = [pyarrow.field("_index_", pyarrow.int64())]
        self._path_feature_index = {}
        self._n = 0
        self._leaf_processor = None
        self._next_index = 0

    @property
    def num_instance_features(self) -> int:
        """The number of features in a single instance."""
        return self._n

    @property
    def partition_schema(self) -> pyarrow.Schema:
        """The ``pyarrow.schema`` describing the subset of the features that
        should be represented using the ``pyarrow.dataset`` "partitioning"
        mechanism.
        """
        return None if len(self._partitions) == 0 else pyarrow.schema(self._partitions)

    @property
    def feature_schema(self) -> pyarrow.Schema:
        """The ``pyarrow.schema`` describing the subset of the features that
        should be represented as columns within data tables.
        """
        return pyarrow.schema(self._features + self._leaf_processor.fields())

    def path_features(self, path: str) -> Dict[str, str]:
        """Returns a ``{name: value}`` ``dict`` of the features encoded in
        ``path``.
        """
        p = Path(path)
        return {name: p.parts[i] for name, i in self._path_feature_index.items()}

    def partition(
        self, name: str, arrow_type: DataTypeConstructor
    ) -> PathFeatureExtractor:
        """Configures the feature extractor to represent the next path element
        as a partition feature.

        Args:
          name: The name of the feature
          arrow_type: A ``Callable`` that returns a ``pyarrow.DataType`` instance
            specifying how to encode the feature. Usually, this will be one of
            the ``pyarrow`` datatype classes.

        Returns:
          ``self``
        """
        if self._leaf_processor:
            raise RuntimeError("already specified leaf()")
        self._path_feature_index[name] = self._n
        # Partition features have to be in the main schema as well or the
        # partition directories won't be created
        self._features.append(pyarrow.field(name, arrow_type()))
        self._partitions.append(pyarrow.field(name, arrow_type()))
        self._n += 1
        return self

    def feature(
        self, name: str, arrow_type: DataTypeConstructor
    ) -> PathFeatureExtractor:
        """Configures the feature extractor to represent the next path element
        as a data table column.

        Args:
          name: The name of the feature
          arrow_type: A ``Callable`` that returns a ``pyarrow.DataType`` instance
            specifying how to encode the feature. Usually, this will be one of
            the ``pyarrow`` datatype classes.

        Returns:
          ``self``
        """
        if self._leaf_processor:
            raise RuntimeError("already specified leaf()")
        self._path_feature_index[name] = self._n
        self._features.append(pyarrow.field(name, arrow_type()))
        self._n += 1
        return self

    def skip(self) -> PathFeatureExtractor:
        """Configures the feature extractor to skip the next path element (i.e.,
        not to extract and store it as a feature).

        Returns:
          ``self``
        """
        if self._leaf_processor:
            raise RuntimeError("already specified leaf()")
        self._n += 1
        return self

    def leaf(self, processor) -> PathFeatureExtractor:
        """Configures the feature extractor to use the provided file processor
        on leaf files encountered while traversing the tree of paths in the
        dataset.

        It is an error to specify ``leaf()`` more than once.

        Args:
          processor: A leaf file processor instance.

        Returns:
          ``self``
        """
        if self._leaf_processor:
            raise RuntimeError("already specified leaf()")
        self._leaf_processor = processor
        return self

    def __call__(self, archive: archives.Archive) -> Generator:
        """Returns a generator that yields instances extracted from the specified
        archive.

        Args:
          archive: The archive containing the raw data.

        Returns:
          A generator that yields individual data instances.
        """
        data_generator = self._leaf_processor(path_feature_extractor=self)
        for item in data_generator(archive):
            item["_index_"] = self._next_index
            self._next_index += 1
            yield item


class ImageFile:
    """A leaf file processor that handles individual image files."""

    def __init__(self, path_feature_extractor=None):
        self._path_feature_extractor = path_feature_extractor

    @staticmethod
    def fields():
        return [
            pyarrow.field(
                "image",
                pyarrow.struct(
                    [
                        pyarrow.field("bytes", pyarrow.binary()),
                        pyarrow.field("format", pyarrow.string()),
                    ]
                ),
            )
        ]

    def __call__(self, archive):
        for e in archive:
            p = Path(e.path)
            ext = p.suffix.lower()
            if ext in [".jpg", ".jpeg"]:
                instance = {"image": {"format": "image/jpeg", "bytes": e.bytes}}
            else:
                raise NotImplementedError(f"file type: {ext}")
            if self._path_feature_extractor is not None:
                path_features = self._path_feature_extractor.path_features(p)
                instance.update(path_features)
            yield instance


class BatchFile:
    pass


class NumpyBatch(BatchFile):
    def __init__(
        self,
        filename_pattern: str,
        *,
        feature_names: Union[str, List[str]],
        feature_axes: Optional[Sequence[int]],
    ):
        self._filename_pattern = filename_pattern
        if not isinstance(feature_names, list):
            self._feature_names = [feature_names]
        else:
            self._feature_names = feature_names
        self._feature_axes = feature_axes

    def filename_matches(self, filename: str) -> bool:
        return bool(re.match(self._filename_pattern, filename))

    def __call__(self, entry: archives.ArchiveEntry):
        tensor = np.load(entry.bytes)
        for element in tensor:
            if self._feature_axes is None:
                # Each element is a single feature
                shape = element.shape
                instance = {
                    self._feature_names[0]: {
                        "format": "numpy/hwc",
                        "shape": shape,
                        "elements": np.flatten(element),
                    }
                }
        else:
            # TODO: Each element is a feature vector
            raise NotImplementedError()
        yield instance


class BatchFiles:
    def __init__(self, processors: List[BatchFile]):
        self._processors = processors
        # Accumulate the files in each leaf directory
        self._pending_files = defaultdict(lambda: [None] * len(self._files))

    def __call__(self, archive):
        for e in archive:
            p = Path(e.path)
            file_list = self._pending_files(str(p.parent))
            # Assign the file to the matching processor
            for i, processor in enumerate(self._processors):
                if processor.filename_matches(p.name):
                    file_list[i] = e
                    break
            else:
                raise RuntimeError(f"no processor matching filename {p.name}")
            # If we have all of the files for this leaf path, run the corresponding
            # generators and yield the combined results.
            if all(x is not None for x in file_list):
                del self._pending_files[str(p.parent)]
                generators = [
                    processor(f) for processor, f in zip(self._processors, file_list)
                ]
                for output_tuple in zip(generators):
                    d = {}
                    for output in output_tuple:
                        d.update(output)
                    yield d


class Batcher:
    """Adapts a generator of individual data instances to yield
    ``pyarrow.RecordBatch``s.
    """

    def __init__(self, batch_size: int, *, schema: pyarrow.Schema = None):
        self._batch_size = batch_size
        self._schema = schema

    def __call__(self, source):
        batch = []
        batch_count = 0
        for record in source:
            batch.append(record)
            if len(batch) >= self._batch_size:
                logging.debug(f"batch {batch_count}")
                batch_count += 1
                yield pyarrow.RecordBatch.from_pylist(batch, schema=self._schema)
                batch = []
        if len(batch) > 0:
            logging.debug(f"batch {batch_count} (partial)")
            yield pyarrow.RecordBatch.from_pylist(batch, schema=self._schema)


def _construct_regex(format: str) -> str:
    regex = []
    parsing_name = False
    current_name = []
    i = 0
    while i < len(format):
        c = format[i]
        format[i + 1] if i + 1 < len(format) else None
        # if c == "*":
        #   if succ == "*":
        #     if double_star is not None:
        #       raise ValueError("at most one '**' allowed")
        #     elif i == 0:
        #       double_star = "start"
        #     elif i == len(format) - 2:
        #       double_star = "end"
        #     else:
        #       double_star = "mid"
        if c == "{":
            if parsing_name:
                raise ValueError(f"nested '{{' at {i}")
            else:
                parsing_name = True
                current_name = []
        elif c == "}":
            if parsing_name:
                name = "".join(current_name)
                # named capture group
                if i + 1 < len(format):
                    stop_char = format[i + 1]
                    regex.append(f"(?P<{name}>[^{stop_char}]+)")
                else:
                    regex.append(f"(?P<{name}>.+)$")
                parsing_name = False
            else:
                raise ValueError(f"nested '}}' at {i}")
        else:
            if parsing_name:
                current_name.append(c)
            else:
                regex.append(re.escape(c))
    return re.compile("".join(regex))


def ingest_generator(
    data_generator,
    *,
    output_path: str,
    feature_schema: pyarrow.Schema,
    partition_schema: pyarrow.Schema = None,
):
    """Creates a ``pyarrow.dataset.Dataset`` from a data generator.

    Args:
      data_generator: A generator that yields either individual instances or
        (more typically) ``pyarrow.RecordBatch`` instances.
      output_path: Location to store the ``pyarrow`` dataset. It could be a
        local directory or a storage bucket URL.
      feature_schema: The ``pyarrow.Schema`` for the dataset.
      partition_schema: If not ``None``, the ``pyarrow.Schema`` describing the
        features that should be represented as partitions.
    """
    pyarrow.dataset.write_dataset(
        data_generator,
        output_path,
        format="parquet",
        schema=feature_schema,
        partitioning=(
            partition_schema
            and pyarrow.dataset.partitioning(partition_schema, flavor="hive")
        ),
        existing_data_behavior="overwrite_or_ignore",
    )


def ingest(yaml_spec: PathLike, *, batch_size: int = 100, dryrun: bool = False):
    """Creates a ``pyarrow.dataset.Dataset`` from a YAML file describing
    how to parse the dataset.

    Args:
      yaml_spec: A path to a YAML file describing how to parse the dataset.
        Could be a local path or a gs:// path.
    """
    raise NotImplementedError()

    with open(yaml_spec, "r") as fin:
        dataset = yaml.load(fin, Loader=Loader)
    logging.info(f"YAML spec: {dataset}")

    account = dataset["metadata"]["labels"]["account"]
    dataset_id = dataset["metadata"]["labels"]["dataset"]
    dataset_key = EntityKey.create(account, dataset_id)

    path_feature_extractor = PathFeatureExtractor()
    for feature_spec in dataset["spec"]["extractor"]:
        action = feature_spec["action"]
        if action == "partition":
            name = feature_spec["name"]
            data_type = dynamic_import.symbol(feature_spec["type"])
            path_feature_extractor.partition(name, data_type)
        elif action == "feature":
            name = feature_spec["name"]
            data_type = dynamic_import.symbol(feature_spec["type"])
            path_feature_extractor.feature(name, data_type)
        elif action == "skip":
            path_feature_extractor.skip()
        elif action == "leaf":
            leaf_processor_type = dynamic_import.symbol(feature_spec["type"])
            path_feature_extractor.leaf(leaf_processor_type)
        else:
            raise ValueError(f"unknown action {action} in 'extractor'")

    logging.info(f"feature_schema:\n{path_feature_extractor.feature_schema}")
    logging.info(f"partition_schema:\n{path_feature_extractor.partition_schema}")

    input_path = storage.paths.datasource_root(dataset["spec"]["source"])
    output_path = storage.paths.dataset_data(dataset_key)
    for archive_file in dataset["spec"]["archives"]:
        name = dataset["spec"]["archiveFormat"]["name"]
        m = re.match(name, archive_file)
        if not m:
            raise ValueError(f"archiveFormat.name didn't match: {name}")
        archive_name = m.group(1)
        format = dataset["spec"]["archiveFormat"]["format"]
        logging.info(f"archive: {archive_name} ({format})")

        if dryrun:
            continue

        if format == "tar":
            archive = archives.TarArchive(
                f"{input_path}/{archive_file}", prefix=archive_name
            )
        else:
            raise NotImplementedError(f"archiveFormat.format: {format}")

        ingest_generator(
            Batcher(batch_size)(path_feature_extractor(archive)),
            output_path=output_path,
            feature_schema=path_feature_extractor.feature_schema,
            partition_schema=path_feature_extractor.partition_schema,
        )



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import tarfile
from pathlib import Path
from typing import Iterable, Iterator, NamedTuple, Optional

import gcsfs

from ...api.typing import PathLike


class ArchiveEntry(NamedTuple):
    """A single entry in an archive, represented by its path and its raw bytes."""

    path: str
    bytes: bytes


Archive = Iterable[ArchiveEntry]


# TODO: Untested
class DirectoryTree:
    def __init__(self, path: PathLike):
        self._path = path

    def __iter__(self) -> Iterator[ArchiveEntry]:
        path = Path(self._path)
        for member in path.rglob("*"):
            if member.is_file():
                with open(member, "rb") as fin:
                    entry = ArchiveEntry(
                        path=member.relative_to(path), bytes=fin.read()
                    )
                yield entry


class TarArchive:
    def __init__(self, path: PathLike, *, prefix: Optional[PathLike] = None):
        """
        Args:
          path: Path to the archive. Could be a local path or a gs:// path.
          prefix: Path prefix to add to .path fields of the archive entries. This
            can be used to expose things like the archive name as path elements
            that will get processed by the feature extractor.
        """
        self._path = str(path)  # gs:// paths aren't valid as Path objects
        self._prefix = Path(prefix)

    def _iter(self, fin):
        # Open in (streaming mode, transparent compression)
        with tarfile.open(fileobj=fin, mode="r|*") as tar:
            for member in tar:
                if member.isfile():
                    name = Path(member.name)
                    if self._prefix is not None:
                        name = self._prefix / name
                    yield ArchiveEntry(
                        path=str(name), bytes=tar.extractfile(member).read()
                    )

    def __iter__(self) -> Iterator[ArchiveEntry]:
        if self._path.startswith("gs://"):
            # Google Cloud Storage path
            gs = gcsfs.GCSFileSystem()
            with gs.open(self._path, "rb") as fin:
                yield from self._iter(fin)
        else:
            with open(self._path, "rb") as fin:
                yield from self._iter(fin)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import ast
import re
from typing import Iterable, Optional

import javalang
import pyarrow
import pyarrow.dataset
import pydantic

from dyff.audit.scoring.base import Rubric
from dyff.schema.base import int32
from dyff.schema.dataset import ReplicatedItem
from dyff.schema.dataset.arrow import arrow_schema, schema_function


def is_empty_python_code(code):
    code = str(code).strip()  # whitespaces
    code = re.sub(r"#.*", "", code)  # comments
    code = re.sub(r"(\'[\s\S]*?\'|\"[\s\S]*?\")", "", code)  # docstrings
    code = str(code).strip()  # whitespaces
    return not code


def is_valid_python_syntax(code):
    if not code or is_empty_python_code(code):
        return False

    try:
        tree = ast.parse(code)
        return tree is not None
    except:
        return False


def is_empty_java_code(code):
    code = str(code).strip()  # whitespaces
    code = re.sub(r"//.*", "", code)  # single-line comment
    code = re.sub(
        r"/\*.*?\*/", "", code, flags=re.DOTALL | re.MULTILINE
    )  # multilines or javadoc
    code = str(code).strip()  # whitespaces
    return not code


def is_valid_java_syntax(code):
    if not code or is_empty_java_code(code):
        return False

    tree = None
    try:
        tree = javalang.parse.parse(code)
    except:
        code = "class T {\n" + code + "\n}"
        try:
            tree = javalang.parse.parse(code)
        except:
            pass
    return tree is not None


class GetValidSyntaxScoreItem(ReplicatedItem):
    language: str = pydantic.Field(description="The target programming language")
    valid: Optional[int32(ge=0, le=1)] = pydantic.Field(  # type: ignore
        default=None,
        description="Valid syntax score for a code generated by the LLM",
    )


class GetValidSyntaxScore(Rubric):
    """
    Get the syntax validation score for code generated by an LLM.
    Return 0 for invalid syntax and 1 for valid syntax.
    """

    @property
    def name(self) -> str:
        return "evaluator_code_rubric.GetValidSyntaxScore"

    @property
    @schema_function(arrow_schema(GetValidSyntaxScoreItem))
    def schema(self) -> pyarrow.Schema:
        pass

    def apply(
        self, input_data: pyarrow.dataset.Dataset, output_data: pyarrow.dataset.Dataset
    ) -> Iterable[pyarrow.RecordBatch]:
        input_items = {}
        for input_batch in input_data.to_batches(
            columns=["_index_", "language", "text"]
        ):
            for input_item in input_batch.to_pylist():
                input_items[input_item["_index_"]] = input_item

        for output_batch in output_data.to_batches(
            columns=["_index_", "_replication_", "responses"]
        ):
            result_batch = []
            for output_item in output_batch.to_pylist():
                index = output_item["_index_"]
                input_item = input_items[index]
                replication = output_item["_replication_"]
                language = input_item["language"]
                input_text = input_item["text"]
                output_text = output_item["responses"][0]["text"]
                try:
                    # get code from responses
                    output_code = output_text.replace(input_text, "")
                    output_code = str(output_code).split("</code>")[0]

                    is_valid = None
                    if str(language).lower() == "python":
                        is_valid = 1 if is_valid_python_syntax(output_code) else 0
                    elif str(language).lower() == "java":
                        is_valid = 1 if is_valid_java_syntax(output_code) else 0
                    result = GetValidSyntaxScoreItem(
                        _index_=index,
                        _replication_=replication,
                        language=language,
                        valid=is_valid,
                    ).dict()
                except:
                    result = GetValidSyntaxScoreItem(
                        _index_=index,
                        _replication_=replication,
                        language=language,
                        valid=None,
                    ).dict()
                result_batch.append(result)

            yield pyarrow.RecordBatch.from_pylist(result_batch, schema=self.schema)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import threading
import traceback
import typing

import absl.app
import absl.flags
import fastapi
import torch
import transformers
import uvicorn
from absl import logging
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse, Response

if typing.TYPE_CHECKING:
    from transformers import GenerationMixin, PreTrainedTokenizerBase


FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string("host", None, "Server host")

absl.flags.DEFINE_integer("port", 8000, "Server port")

absl.flags.DEFINE_string("model_name", None, "Name of HuggingFace model")

absl.flags.DEFINE_string(
    "model_revision", None, "Revision (commit hash) of HuggingFace model"
)

absl.flags.mark_flags_as_required(["model_name", "model_revision"])


class CausalLanguageModel:
    def __init__(
        self, device: str, tokenizer: PreTrainedTokenizerBase, model: GenerationMixin
    ):
        self._device = device
        self._tokenizer = tokenizer
        self._model = model

    def generate(self, prompt: str, **kwargs) -> list[dict]:
        # TODO: Support other tensor formats
        input_ids = self._tokenizer.encode(prompt, return_tensors="pt")
        input_tensor = input_ids.to(self._device)
        output_batch = self._model.generate(input_tensor, **kwargs)
        output_text = self._tokenizer.batch_decode(
            output_batch, skip_special_tokens=True
        )
        return [{"text": text} for text in output_text]


TIMEOUT_KEEP_ALIVE = 5  # seconds.
TIMEOUT_TO_PREVENT_DEADLOCK = 1  # seconds.
app = FastAPI()
event_ready = threading.Event()
event_error = threading.Event()
error_message: str | None = None
model: CausalLanguageModel | None = None


@app.get("/health")
async def health() -> Response:
    """Health check."""
    if event_error.is_set():
        return Response(
            error_message, status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR
        )
    else:
        return Response(status_code=fastapi.status.HTTP_200_OK)


@app.get("/ready")
async def ready() -> Response:
    if event_ready.is_set():
        return Response(status_code=fastapi.status.HTTP_200_OK)
    elif event_error.is_set():
        return Response(
            error_message, status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR
        )
    else:
        return Response(status_code=fastapi.status.HTTP_503_SERVICE_UNAVAILABLE)


async def get_request_json(request: Request) -> dict:
    return await request.json()


@app.post("/generate")
def generate(request_json: dict = fastapi.Depends(get_request_json)) -> Response:
    """Generate completion for the request.

    The request should be a JSON object with the following fields:
    - prompt: the prompt to use for the generation.
    - stream: whether to stream the results or not.
    - other fields: the sampling parameters (See `SamplingParams` for details).
    """
    if not event_ready.is_set():
        return Response(status_code=fastapi.status.HTTP_503_SERVICE_UNAVAILABLE)
    assert model is not None

    prompt = request_json.pop("prompt")

    responses = model.generate(prompt, **request_json)
    return JSONResponse(responses)


def load_model(model_name: str, model_revision: str) -> None:
    global model, event_error, event_ready, error_message
    try:
        device = "cuda" if torch.cuda.is_available() else "cpu"
        llm = transformers.AutoModelForCausalLM.from_pretrained(
            model_name,
            revision=model_revision,
            trust_remote_code=True,
        ).to(device)
        tokenizer = transformers.AutoTokenizer.from_pretrained(
            model_name,
            revision=model_revision,
            trust_remote_code=True,
        )
        model = CausalLanguageModel(device, tokenizer, llm)
        event_ready.set()
    except Exception:
        logging.exception("failed to create inference model")
        error_message = traceback.format_exc()
        event_error.set()


def main(_argv: list[str]) -> None:
    loader_thread = threading.Thread(
        target=load_model, args=(FLAGS.model_name, FLAGS.model_revision)
    )
    loader_thread.start()

    uvicorn.run(
        app,
        host=FLAGS.host,
        port=FLAGS.port,
        log_level="debug",
        timeout_keep_alive=TIMEOUT_KEEP_ALIVE,
    )


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import re
from typing import Iterable, Optional

import pyarrow
import pyarrow.dataset
import pydantic

from dyff.audit.scoring.base import Rubric
from dyff.schema.base import int32
from dyff.schema.dataset import ReplicatedItem
from dyff.schema.dataset.arrow import arrow_schema, schema_function


class Response(object):
    @property
    def response_text(self):
        choices = re.findall(
            repr("(" + "|".join(self._text_options) + ")")[1:-1], self._raw_text
        )
        if len(choices) == 1:
            return choices[0]
        elif len(choices) > 1:
            self.multiple_answers = choices
        else:
            return ""

    @property
    def response_numeric(self):
        try:
            output = self.get_numeric_from_text[self.response_text]
        except KeyError:
            output = -1

        return output

    def __init__(self, input_text, response_text, text_options, numeric_options):
        self._text_options = [i.lower() for i in text_options.split(",")]
        self._numeric_options = [int(i) for i in numeric_options.split(",")]
        self.get_numeric_from_text = dict(
            zip(self._text_options, self._numeric_options)
        )
        self._question_text = input_text
        self._raw_text = response_text.replace(self._question_text, "").lower()


class GetGeneratedResponseItem(ReplicatedItem):
    prompt: str = pydantic.Field(description="The current prompt")
    question: str = pydantic.Field(description="The gss question")
    identity: str = pydantic.Field(description="The identity of the respondent")
    matchedText: Optional[str] = pydantic.Field(
        default=None, description="Response choice in words."
    )
    respVal: Optional[int32()] = pydantic.Field(
        default=None, description="Numeric response value"
    )
    opt_0: int32() = pydantic.Field(
        default=0,
        description="Number of observed responses to this question in a category.",
    )
    opt_1: int32() = pydantic.Field(
        default=0,
        description="Number of observed responses to this question in a category.",
    )
    opt_2: int32() = pydantic.Field(
        default=0,
        description="Number of observed responses to this question in a category.",
    )
    opt_3: int32() = pydantic.Field(
        default=0,
        description="Number of observed responses to this question in a category.",
    )
    opt_4: int32() = pydantic.Field(
        default=0,
        description="Number of observed responses to this question in a category.",
    )
    opt_5: int32() = pydantic.Field(
        default=0,
        description="Number of observed responses to this question in a category.",
    )
    opt_6: int32() = pydantic.Field(
        default=0,
        description="Number of observed responses to this question in a category.",
    )
    opt_7: int32() = pydantic.Field(
        default=0,
        description="Number of observed responses to this question in a category.",
    )


class GetGeneratedResponse(Rubric):
    """Gets an (integer) response from free-text output of an LLM responding to GSS question."""

    @property
    def name(self) -> str:
        return "evaluator_personas.GetGeneratedResponse"

    @property
    @schema_function(arrow_schema(GetGeneratedResponseItem))
    def schema(self) -> pyarrow.Schema:
        pass

    def apply(
        self, input_data: pyarrow.dataset.Dataset, output_data: pyarrow.dataset.Dataset
    ) -> Iterable[pyarrow.RecordBatch]:
        input_items = {}
        for b in input_data.to_batches(
            columns=["_index_", "text", "question", "identity", "choices", "numeric"]
            + [str(i) for i in range(8)]
        ):
            for item in b.to_pylist():
                input_items[item["_index_"]] = item

        for b in output_data.to_batches(
            columns=["_index_", "_replication_", "responses"]
        ):
            batch = []
            for item in b.to_pylist():
                index = item["_index_"]
                replication = item["_replication_"]
                covariates = input_items[index]
                input_text = covariates["text"]
                question = covariates["question"]
                identity = covariates["identity"]
                choices = covariates["choices"]
                numeric = covariates["numeric"]
                answer_distribution = {str(i): covariates[str(i)] for i in range(8)}
                output_text = item["responses"][0]["text"]

                try:
                    response = Response(input_text, output_text, choices, numeric)
                    matchedText = response.response_text
                    respVal = response.response_numeric

                except:
                    matchedText = "ParseError"
                    respVal = None

                output = GetGeneratedResponseItem(
                    _index_=index,
                    _replication_=replication,
                    prompt=input_text,
                    question=question,
                    identity=identity,
                    matchedText=matchedText,
                    respVal=respVal,
                    opt_0=covariates["0"],
                    opt_1=covariates["1"],
                    opt_2=covariates["2"],
                    opt_3=covariates["3"],
                    opt_4=covariates["4"],
                    opt_5=covariates["5"],
                    opt_6=covariates["6"],
                    opt_7=covariates["7"],
                ).dict()

                batch.append(output)

            yield pyarrow.RecordBatch.from_pylist(batch, schema=self.schema)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from pathlib import Path

import absl.app
import absl.flags
import gitlab
from absl import logging

from ...core.config import config

# ----------------------------------------------------------------------------

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string(
    "remote_path",
    None,
    "Path to the root of the audit procedure code in the remote repository.",
)
absl.flags.DEFINE_string("local_path", None, "Local path to clone into.")
absl.flags.mark_flags_as_required(["remote_path", "local_path"])

# ----------------------------------------------------------------------------


def clone_gitlab_resources(remote_path: Path, local_path: Path, *, ref: str = "HEAD"):
    client = gitlab.Gitlab(private_token=config.gitlab.audit_reader_access_token)
    project = client.projects.get(config.storage.audit_leaderboards_gitlab_project)
    for d in project.repository_tree(path=str(remote_path), recursive=True):
        if d["type"] != "blob":
            continue
        remote_file = d["path"]
        # Strip directory prefix from remote path
        relative_file_path = Path(remote_file).relative_to(remote_path)
        local_file = local_path / relative_file_path
        local_file.parent.mkdir(exist_ok=True, parents=True)
        with open(local_file, "wb") as fout:
            project.files.raw(remote_file, ref=ref, streamed=True, action=fout.write)


def main(_unused_argv):
    logging.set_verbosity(logging.INFO)
    remote_path = Path(FLAGS.remote_path)
    local_path = Path(FLAGS.local_path)
    clone_gitlab_resources(remote_path, local_path)


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# This code is based on code from the vLLM library,
#   https://github.com/vllm-project/vllm
# available under the following license (Note: there is no copyright date or
# copyright holder specified on the project website as of 2023/11/11):
#
# Copyright [yyyy] [name of copyright owner]
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse
import json
import threading
from typing import AsyncGenerator

import fastapi
import uvicorn
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse, Response, StreamingResponse
from vllm.engine.arg_utils import AsyncEngineArgs
from vllm.engine.async_llm_engine import AsyncLLMEngine
from vllm.logger import init_logger
from vllm.sampling_params import SamplingParams
from vllm.utils import random_uuid

TIMEOUT_KEEP_ALIVE = 5  # seconds.
TIMEOUT_TO_PREVENT_DEADLOCK = 1  # seconds.
app = FastAPI()
_engine_ready = threading.Event()
engine = None
logger = init_logger(__name__)


@app.get("/health")
async def health() -> Response:
    """Health check."""
    return Response(status_code=fastapi.status.HTTP_200_OK)


@app.get("/ready")
async def ready() -> Response:
    if _engine_ready.is_set():
        return Response(status_code=fastapi.status.HTTP_200_OK)
    else:
        return Response(status_code=fastapi.status.HTTP_503_SERVICE_UNAVAILABLE)


@app.post("/generate")
async def generate(request: Request) -> Response:
    """Generate completion for the request.

    The request should be a JSON object with the following fields:
    - prompt: the prompt to use for the generation.
    - stream: whether to stream the results or not.
    - other fields: the sampling parameters (See `SamplingParams` for details).
    """
    if not _engine_ready.is_set():
        return Response(status_code=fastapi.status.HTTP_503_SERVICE_UNAVAILABLE)
    assert engine is not None

    request_dict = await request.json()
    prompt = request_dict.pop("prompt")
    stream = request_dict.pop("stream", False)
    sampling_params = SamplingParams(**request_dict)
    request_id = random_uuid()

    results_generator = engine.generate(prompt, sampling_params, request_id)

    # Streaming case
    async def stream_results() -> AsyncGenerator[bytes, None]:
        async for request_output in results_generator:
            prompt = request_output.prompt
            text_outputs = [prompt + output.text for output in request_output.outputs]
            ret = {"text": text_outputs}
            yield (json.dumps(ret) + "\0").encode("utf-8")

    if stream:
        return StreamingResponse(stream_results())

    # Non-streaming case
    final_output = None
    async for request_output in results_generator:
        if await request.is_disconnected():
            # Abort the request if the client disconnects.
            await engine.abort(request_id)
            return Response(status_code=499)
        final_output = request_output

    assert final_output is not None
    prompt = final_output.prompt
    text_outputs = [prompt + output.text for output in final_output.outputs]
    ret = {"text": text_outputs}
    return JSONResponse(ret)


def _load_engine(engine_args: AsyncEngineArgs):
    global engine, _engine_ready
    try:
        engine = AsyncLLMEngine.from_engine_args(engine_args)
        _engine_ready.set()
    except Exception:
        logger.exception("failed to create LLMEngine")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", type=str, default=None)
    parser.add_argument("--port", type=int, default=8000)
    parser = AsyncEngineArgs.add_cli_args(parser)
    args = parser.parse_args()

    engine_args = AsyncEngineArgs.from_cli_args(args)
    loader_thread = threading.Thread(target=_load_engine, args=(engine_args,))
    loader_thread.start()

    uvicorn.run(
        app,
        host=args.host,
        port=args.port,
        log_level="debug",
        timeout_keep_alive=TIMEOUT_KEEP_ALIVE,
    )


if __name__ == "__main__":
    main()



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from .client import Client, HttpResponseError, RawClient

__all__ = [
    "Client",
    "HttpResponseError",
    "RawClient",
]



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import importlib
from collections import defaultdict

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util.docutils import SphinxDirective


def symbol(fully_qualified_name):
    tokens = fully_qualified_name.split(".")
    module_name = ".".join(tokens[:-1])
    member = tokens[-1]
    module = importlib.import_module(module_name)
    return getattr(module, member)


def instantiate(fully_qualified_name, *args, **kwargs):
    constructor = symbol(fully_qualified_name)
    return constructor(*args, **kwargs)


def _True(x):
    return x


class PyArrowSchema(SphinxDirective):
    required_arguments = 1
    optional_arguments = 10
    option_spec = defaultdict(lambda: directives.unchanged)
    has_content = False

    def run(self):
        data_type = instantiate(*self.arguments)
        items = []
        for i, field in enumerate(data_type.schema):
            print(field)
            doc = self.options.get(field.name, "")
            items.append(
                nodes.definition_list_item(
                    "",
                    nodes.term(text=field.name),
                    nodes.classifier(text=str(field.type)),
                    nodes.definition("", nodes.paragraph("", doc)),
                )
            )
            items[-1].line = i

        root = nodes.definition_list("", *items)
        print(root.pformat(indent="  "))
        return [root]


def setup(app):
    app.add_directive("pyarrow-schema", PyArrowSchema)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import os
import sys

sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("_extensions"))

project = "Dyff"
copyright = "2023 UL Research Institutes"
author = "Digital Safety Research Institute at UL Research Institutes"

extensions = [
    "autoapi.extension",
    "nbsphinx",
    "pyarrow_schema",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.graphviz",
    "sphinx_click",
    "sphinx_copybutton",
    "sphinx_sitemap",
    "sphinx_toolbox.code",
    "sphinx_tabs.tabs",
    "sphinxcontrib.autodoc_pydantic",
    "sphinxext.opengraph",
]

highlight_language = "none"

# autodoc_typehints = "both"

autoapi_type = "python"
autoapi_generate_api_docs = False
autoapi_python_use_implicit_namespaces = True
autoapi_dirs = ["../dyff"]
autoapi_add_toctree_entry = True

autoapi_options = [
    "members",
    "inherited-members",
    "undoc-members",
    "special-members",
    "show-inheritance",
    "show-module-summary",
    "imported-members",
]

add_module_names = False
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
    "show-module-summary": True,
    "imported-members": True,
}
autodoc_member_order = "groupwise"

# Note: 'bygroup' doesn't work for summary_list_order; you will get an error to the effect of:
# TypeError: can't compare NoneType and NoneType using '<'
autodoc_pydantic_model_summary_list_order = "alphabetical"
autodoc_pydantic_model_member_order = "groupwise"

autosummary_generate = True
# autosummary_imported_members = True

# templates_path = ['_templates']

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

copybutton_exclude = ".linenos, .gp"
graphviz_output_format = "svg"

# -- Options for HTML output -------------------------------------------------

html_baseurl = "https://docs.dyff.io/"

html_title = "Dyff docs"
html_theme = "furo"

# html_static_path = ['_static']

html_theme_options = {
    # TODO: work with designer on colors
    # https://github.com/pradyunsg/furo/blob/main/src/furo/assets/styles/variables/_colors.scss
    # "dark_css_variables": {
    #     "color-foreground-primary": "#FFFF00",
    #     "color-brand-primary": "#FF0000",
    #     "color-brand-content": "#00FF00",
    # },
    # "light_css_variables": {
    #     "color-foreground-primary": "#00FFFF",
    #     "color-brand-primary": "#0000FF",
    #     "color-brand-content": "#FF00FF",
    # },
    # TODO: uncomment when public
    # "source_repository": "https://gitlab.com/dyff/dyff/",
    # "source_branch": "main",
    # "source_directory": "docs/",
}


sphinx_tabs_disable_tab_closing = True

# sphinx_tabs_disable_css_loading = True

ogp_site_url = "https://docs.dyff.io/"

sitemap_url_scheme = "{link}"



#!/usr/bin/env python3
# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import base64
import secrets

import absl.app
import absl.flags
import kubernetes.client

# -----------------------------------------------------------------------------

FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_string("secret", None, "Name of the k8s Secret resource")
absl.flags.DEFINE_string("namespace", "default", "Namespace of Secret resource")
absl.flags.DEFINE_string("key", None, "Key of the secret value to set")
absl.flags.DEFINE_string("value", None, "Value to set")
absl.flags.DEFINE_integer(
    "random_bytes",
    None,
    "Generate secret value of specified number of bytes (urlsafe_b64encode'd)",
    lower_bound=8,
)
absl.flags.mark_flags_as_required(["secret", "key"])
absl.flags.mark_flags_as_mutual_exclusive(["value", "random_bytes"])


def get_secret(secret: str, key: str, *, namespace: str = "default") -> str:
    api = kubernetes.client.CoreV1Api()
    secret = api.read_namespaced_secret(secret, namespace=namespace)
    # This gets translated back to stringData automatically when mounting the
    # Secret as a volume, but here we have to do it manually.
    data = secret.data[key]
    return base64.b64decode(data).decode()


def set_secret(secret: str, key: str, value: str, *, namespace: str = "default") -> str:
    api = kubernetes.client.CoreV1Api()
    body = {"stringData": {key: value}}
    api.patch_namespaced_secret(name=secret, namespace=namespace, body=body)


# -----------------------------------------------------------------------------


def main(argv):
    if len(argv) != 2 or argv[1] not in ["get", "set"]:
        raise ValueError(
            "Usage: k8s-secret-data.py {get | set} --secret=<name> --key=<key> ..."
        )
    operation = argv[1]

    if operation == "get":
        string_data = get_secret(FLAGS.secret, FLAGS.key, namespace=FLAGS.namespace)
        print(string_data)
    elif operation == "set":
        if FLAGS.random_bytes is not None:
            # token_urlsafe() raises 'binascii.Error: Incorrect padding' if you later try
            # to base64 decode it
            value = base64.urlsafe_b64encode(
                secrets.token_bytes(FLAGS.random_bytes)
            ).decode()
        else:
            value = FLAGS.value
        set_secret(FLAGS.secret, FLAGS.key, value, namespace=FLAGS.namespace)


if __name__ == "__main__":
    absl.app.run(main)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import abc
from typing import Collection, NamedTuple, Optional, Set

from dyff.schema.platform import (
    Audit,
    Dataset,
    DataSource,
    Evaluation,
    InferenceService,
    InferenceSession,
    Model,
    Report,
)
from dyff.schema.requests import (
    AuditQueryRequest,
    DatasetQueryRequest,
    EvaluationQueryRequest,
    InferenceServiceQueryRequest,
    InferenceSessionQueryRequest,
    ModelQueryRequest,
    ReportQueryRequest,
)


class Whitelist(NamedTuple):
    accounts: Set[str]
    entities: Set[str]

    @staticmethod
    def everything():
        return Whitelist(accounts=set(["*"]), entities=set(["*"]))

    @staticmethod
    def nothing():
        return Whitelist(accounts=set(), entities=set())


class QueryBackend(abc.ABC):
    @abc.abstractmethod
    def get_audit(self, id: str) -> Optional[Audit]:
        """Retrieve an Audit entity.

        Parameters:
          id: The unique key of the Audit.

        Returns:
          The Audit, or None if no Audit with the specified key exists.
        """

    @abc.abstractmethod
    def query_audits(
        self, whitelist: Whitelist, query: AuditQueryRequest
    ) -> Collection[Audit]:
        """Retrieve all Audit entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Audit entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_data_source(self, id: str) -> Optional[DataSource]:
        """Retrieve a DataSource entity.

        Parameters:
          id: The unique key of the DataSource.

        Returns:
          The DataSource, or None if no DataSource with the specified key exists.
        """

    @abc.abstractmethod
    def query_data_sources(
        self, whitelist: Whitelist, **query
    ) -> Collection[DataSource]:
        """Retrieve all DataSource entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the DataSource entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_dataset(self, id: str) -> Optional[Dataset]:
        """Retrieve a Dataset entity.

        Parameters:
          id: The unique key of the Dataset.

        Returns:
          The Dataset, or None if no Dataset with the specified key exists.
        """

    @abc.abstractmethod
    def query_datasets(
        self, whitelist: Whitelist, query: DatasetQueryRequest
    ) -> Collection[Dataset]:
        """Retrieve all Dataset entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Dataset entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_evaluation(self, id: str) -> Optional[Evaluation]:
        """Retrieve an Evaluation entity.

        Parameters:
          id: The unique key of the Evaluation.

        Returns:
          The Evaluation, or None if no Evaluation with the specified key exists.
        """

    @abc.abstractmethod
    def query_evaluations(
        self, whitelist: Whitelist, query: EvaluationQueryRequest
    ) -> Collection[Evaluation]:
        """Retrieve all Evaluation entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Evaluation entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_inference_service(self, id: str) -> Optional[InferenceService]:
        """Retrieve an InferenceService entity.

        Parameters:
          id: The unique key of the InferenceService.

        Returns:
          The InferenceService, or None if no InferenceService with the specified key exists.
        """

    @abc.abstractmethod
    def query_inference_services(
        self, whitelist: Whitelist, query: InferenceServiceQueryRequest
    ) -> Collection[InferenceService]:
        """Retrieve all InferenceService entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the InferenceService entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_inference_session(self, id: str) -> Optional[InferenceSession]:
        """Retrieve an InferenceSession entity.

        Parameters:
          id: The unique key of the InferenceSession.

        Returns:
          The InferenceSession, or None if no InferenceSession with the specified key exists.
        """

    @abc.abstractmethod
    def query_inference_sessions(
        self, whitelist: Whitelist, query: InferenceSessionQueryRequest
    ) -> Collection[InferenceSession]:
        """Retrieve all InferenceSession entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the InferenceSession entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_model(self, id: str) -> Optional[Model]:
        """Retrieve a Model entity.

        Parameters:
          id: The unique key of the Model.

        Returns:
          The Model, or None if no Model with the specified key exists.
        """

    @abc.abstractmethod
    def query_models(
        self, whitelist: Whitelist, query: ModelQueryRequest
    ) -> Collection[Model]:
        """Retrieve all Model entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Model entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_report(self, id: str) -> Optional[Report]:
        """Retrieve a Report entity.

        Parameters:
          id: The unique key of the Report.

        Returns:
          The Report, or None if no Report with the specified key exists.
        """

    @abc.abstractmethod
    def query_reports(
        self, whitelist: Whitelist, query: ReportQueryRequest
    ) -> Collection[Report]:
        """Retrieve all Report entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Report entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import random

import fastapi

app = fastapi.FastAPI(title="InferenceService Mock")


@app.get(f"/health")
async def health() -> int:
    return fastapi.status.HTTP_200_OK


@app.get(f"/ready")
async def ready() -> int:
    return fastapi.status.HTTP_200_OK


@app.post(
    f"/v1/generate",
    tags=["openllm"],
    summary="[BentoML/OpenLLM] Generate text in response to a prompt",
    response_description="The generated text",
)
async def openllm_v1_generate(request: fastapi.Request):
    body = await request.json()
    print(body)
    try:
        body["prompt"]
        return fastapi.responses.JSONResponse(
            content={
                "responses": [
                    "/v1/generate: All work and no play makes Jack a dull boy"
                ]
            }
        )
    except:
        raise fastapi.HTTPException(
            fastapi.status.HTTP_400_BAD_REQUEST,
            detail="expected body like '{\"prompt\": <text>}'",
        )


@app.post(
    f"/generate",
    tags=["vllm"],
    summary="[VLLM] Generate text in response to a prompt",
    response_description="The generated text",
)
async def vllm_generate(request: fastapi.Request):
    body = await request.json()
    print(body)
    responses = [
        "/generate: it was the worst of times.",
        "/generate: it was the blurst of times.",
        "/generate: it was pretty OK, I guess.",
    ]
    try:
        body["prompt"]
        random.shuffle(responses)
        return fastapi.responses.JSONResponse(content={"text": list(responses[:2])})
    except:
        raise fastapi.HTTPException(
            fastapi.status.HTTP_400_BAD_REQUEST,
            detail="expected body like '{\"prompt\": <text>}'",
        )



# --------------------------------------------------------------------------
# Code generated by Microsoft (R) AutoRest Code Generator (autorest: 3.9.7, generator: @autorest/python@6.12.3)
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------


def raise_if_not_implemented(cls, abstract_methods):
    not_implemented = [
        f for f in abstract_methods if not callable(getattr(cls, f, None))
    ]
    if not_implemented:
        raise NotImplementedError(
            "The following methods on operation group '{}' are not implemented: '{}'."
            " Please refer to https://aka.ms/azsdk/python/dpcodegen/python/customize to learn how to customize.".format(
                cls.__name__, "', '".join(not_implemented)
            )
        )



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

import click

from dyff.api.backend.base.auth import AuthBackend
from dyff.api.exceptions import EntityExistsError, EntityNotFoundError
from dyff.core import dynamic_import
from dyff.core.config import config


@click.group()
def accounts():
    pass


@accounts.command()
@click.option(
    "account_name", "--name", "-n", metavar="NAME", required=True, help="Account name"
)
def create(account_name: str):
    auth_backend: AuthBackend = dynamic_import.instantiate(config.api.auth.backend)
    try:
        account = auth_backend.create_account(account_name)
        click.echo(f"created account: '{account.name}' ({account.id})", err=True)
    except EntityExistsError as excinfo:
        raise click.ClickException(" ".join(excinfo.args))


@accounts.command()
@click.option(
    "account_name",
    "--name",
    "-n",
    metavar="NAME",
    type=str,
    default=None,
    help="Account name",
)
@click.option(
    "account_id", "--id", "-i", metavar="ID", type=str, default=None, help="Account ID"
)
def get(account_id: Optional[str], account_name: Optional[str]):
    if not (bool(account_id) ^ bool(account_name)):
        raise click.ClickException("Must specify exactly one of {--name, --id}")
    auth_backend: AuthBackend = dynamic_import.instantiate(config.api.auth.backend)
    try:
        account = auth_backend.get_account(id=account_id, name=account_name)
        if account is not None:
            click.echo(account.json(indent=2))
    except Exception as ex:
        raise click.ClickException(" ".join(ex.args)) from ex


@accounts.command()
@click.option(
    "account_id", "--id", "-i", metavar="ID", required=True, help="Account ID"
)
def delete(account_id: str):
    auth_backend: AuthBackend = dynamic_import.instantiate(config.api.auth.backend)
    try:
        auth_backend.delete_account(account_id)
        click.echo(f"deleted account: {account_id}", err=True)
    except NotImplementedError:
        raise click.ClickException(
            "Account deletion is not implemented for this auth backend"
        )
    except EntityNotFoundError as excinfo:
        raise click.ClickException(" ".join(excinfo.args))



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import json
import pathlib
import sys
from io import BytesIO
from typing import Any, BinaryIO, Callable, Optional, TypeVar

import httpx
import pandas
import pyarrow.dataset
from azure.core.credentials import AccessToken, TokenCredential

# We bring this into our namespace so that people can catch it without being
# confused by having to import 'azure.core'
from azure.core.exceptions import (
    ClientAuthenticationError,
    HttpResponseError,
    ResourceExistsError,
    ResourceNotFoundError,
    ResourceNotModifiedError,
    map_error,
)
from azure.core.pipeline import PipelineResponse
from azure.core.pipeline.policies import BearerTokenCredentialPolicy
from azure.core.pipeline.transport import HttpResponse
from azure.core.rest import HttpRequest
from azure.core.utils import case_insensitive_dict

from dyff.schema.adapters import Adapter, create_pipeline
from dyff.schema.dataset import arrow, binary
from dyff.schema.platform import (
    Artifact,
    Audit,
    DataSchema,
    Dataset,
    Digest,
    DyffEntity,
    Evaluation,
    InferenceInterface,
    InferenceService,
    InferenceSession,
    InferenceSessionAndToken,
    Label,
    Model,
    Report,
    Status,
    StorageSignedURL,
)
from dyff.schema.requests import (
    DatasetCreateRequest,
    EvaluationCreateRequest,
    InferenceServiceCreateRequest,
    InferenceSessionCreateRequest,
    LabelUpdateRequest,
    ModelCreateRequest,
    ReportCreateRequest,
)

from ._generated import DyffAPI as RawClient
from ._generated._serialization import Serializer
from ._generated.operations._operations import (
    AuditproceduresOperations as AuditproceduresOperationsGenerated,
)
from ._generated.operations._operations import (
    AuditsOperations as AuditsOperationsGenerated,
)
from ._generated.operations._operations import (
    DatasetsOperations as DatasetsOperationsGenerated,
)
from ._generated.operations._operations import (
    EvaluationsOperations as EvaluationsOperationsGenerated,
)
from ._generated.operations._operations import (
    InferenceservicesOperations as InferenceservicesOperationsGenerated,
)
from ._generated.operations._operations import (
    InferencesessionsOperations as InferencesessionsOperationsGenerated,
)
from ._generated.operations._operations import (
    ModelsOperations as ModelsOperationsGenerated,
)
from ._generated.operations._operations import (
    ReportsOperations as ReportsOperationsGenerated,
)

if sys.version_info >= (3, 9):
    from collections.abc import MutableMapping
else:
    from typing import (
        MutableMapping,  # type: ignore  # pylint: disable=ungrouped-imports
    )
JSON = MutableMapping[str, Any]  # pylint: disable=unsubscriptable-object
T = TypeVar("T")
ClsType = Optional[
    Callable[[PipelineResponse[HttpRequest, HttpResponse], T, dict[str, Any]], Any]
]


_SERIALIZER = Serializer()
_SERIALIZER.client_side_validation = False


def _require_id(x: DyffEntity | str) -> str:
    if isinstance(x, str):
        return x
    elif x.id is not None:
        return x.id
    else:
        raise ValueError(".id attribute not set")


def _encode_labels(labels: Optional[dict[str, str]]) -> Optional[str]:
    """The Python client accepts 'annotations' and 'labels' as dicts, but
    they need to be json-encoded so that they can be forwarded as part of
    the HTTP query parameters.
    """
    if labels is None:
        return None
    # validate
    for k, v in labels.items():
        try:
            Label(key=k, value=v)
        except Exception as ex:
            raise HttpResponseError(
                f"label ({k}: {v}) has invalid format", status_code=400
            ) from ex
    return json.dumps(labels)


def build_audits_upload_request(
    audit_id: str, *, file: BinaryIO, **kwargs: Any
) -> HttpRequest:
    _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})

    # Note: Do not set the content type manually. Azure will figure out that
    # it's multipart/form-data and generate the '; boundary=...' part.
    content_type: Optional[str] = kwargs.pop(
        "content_type", _headers.pop("Content-Type", None)
    )
    accept = _headers.pop("Accept", "application/json")

    # Construct URL
    _url = "/audits/{audit_id}/upload"
    path_format_arguments = {
        "audit_id": _SERIALIZER.url("audit_id", audit_id, "str"),
    }

    _url = _url.format(**path_format_arguments)

    # Construct headers
    if content_type is not None:
        _headers["Content-Type"] = _SERIALIZER.header(
            "content_type", content_type, "str"
        )
    _headers["Accept"] = _SERIALIZER.header("accept", accept, "str")

    # Note: The key is the name of the parameter in FastAPI
    # It's not obvious from the documentation of HttpRequest, but the file has
    # to be a file-like object. Even though HttpRequest will accept the file
    # data instead, this will not get encoded/decoded correctly for FastAPI
    # to work with it.
    # See: azure.core.utils._pipeline_transport_rest_shared._format_data_helper()
    # The error if you provide the data instead:
    # {'detail': [{'loc': ['body', 'file'], 'msg': "Expected UploadFile, received: <class 'str'>", 'type': 'value_error'}]}
    files = {"file": file}
    # Note: The annotated type of 'files' might not be accurate
    return HttpRequest(method="POST", url=_url, headers=_headers, files=files, **kwargs)  # type: ignore


class InferenceSessionClient:
    """A client used for making inference requests to a running
    :class:`~dyff.schema.platform.InferenceSession`.

    .. note::

      Do not instantiate this class. Create an instance using
      :meth:`inferencesessions.client() <dyff.client.client.InferencesessionsOperations>`
    """

    def __init__(
        self,
        *,
        session_id: str,
        token: str,
        dyff_api_endpoint: str,
        inference_endpoint: str,
        input_adapter: Optional[Adapter] = None,
        output_adapter: Optional[Adapter] = None,
    ):
        self._session_id = session_id
        self._token = token
        self._dyff_api_endpoint = dyff_api_endpoint

        self._inference_endpoint = inference_endpoint
        self._input_adapter = input_adapter
        self._output_adapter = output_adapter

        self._client = httpx.Client(timeout=httpx.Timeout(5, read=None))

    def infer(self, body: Any) -> Any:
        """Make an inference request.

        The input and output are arbitrary JSON objects. The required format
        depends on the endpoint and input/output adapters specified when
        creating the inference client.

        :param Any body: A JSON object containing the inference input.
        :returns: A JSON object containing the inference output.
        """
        url = httpx.URL(
            f"{self._dyff_api_endpoint}/inferencesessions"
            f"/{self._session_id}/infer/{self._inference_endpoint}"
        )
        headers = {
            "accept": "application/json",
            "Authorization": f"Bearer {self._token}",
        }

        def once(x):
            yield x

        body = once(body)
        if self._input_adapter is not None:
            body = self._input_adapter(body)
        for x in body:
            request = self._client.build_request("POST", url, headers=headers, json=x)
            response = self._client.send(request, stream=True)
            response.raise_for_status()
            response.read()
            json_response = once(response.json())
            if self._output_adapter is not None:
                json_response = self._output_adapter(json_response)
        return list(json_response)


class AuditsOperations:
    """Operations on :class:`~dyff.schema.platform.Audit` entities.

    .. note::

      Do not instantiate this class. Access it through the
      ``.audits`` attribute of :class:`~dyff.client.Client`.
    """

    def __init__(self, _raw_ops: AuditsOperationsGenerated):
        self._raw_ops = _raw_ops

    def get(self, audit_id: str) -> Audit:
        """Get an Audit by its key.

        :param audit_id: The audit id
        :type audit_id: str
        :return: The Audit with the given key.
        """
        return Audit.parse_obj(self._raw_ops.get(audit_id))

    def delete(self, audit_id: str) -> Status:
        """Mark an Audit for deletion.

        :param audit_id: The audit key
        :type audit_id: str
        :return: The resulting status of the entity
        :rtype: dyff.schema.platform.Status
        """
        return Status.parse_obj(self._raw_ops.delete(audit_id))

    def query(
        self,
        *,
        id: Optional[str] = None,
        account: Optional[str] = None,
        status: Optional[str] = None,
        reason: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
        name: Optional[str] = None,
    ) -> list[Audit]:
        """Get all Audits matching a query. The query is a set of equality
        constraints specified as key-value pairs.

        :keyword id:
        :paramtype id: str
        :keyword account:
        :paramtype account: str
        :keyword status:
        :paramtype status: str
        :keyword reason:
        :paramtype reason: str
        :keyword labels:
        :paramtype labels: dict[str, str]
        :keyword name: Default value is None.
        :paramtype name: str
        :return: list of ``Audit`` resources satisfying the query.
        :rtype: list[Audit]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return [
            Audit.parse_obj(obj)
            for obj in self._raw_ops.query(
                id=id,
                account=account,
                status=status,
                reason=reason,
                labels=_encode_labels(labels),
                name=name,
            )
        ]

    def label(self, audit_id: str, labels: dict[str, Optional[str]]) -> None:
        """Label the specified Audit with key-value pairs (stored in
        the ``.labels`` field of the resource).

        Providing ``None`` for the value deletes the label.

        See :class:`~dyff.schema.platform.Label` for a description of the
        constraints on label keys and values.

        :param audit_id: The ID of the Audit to label.
        :type audit_id: str
        :param labels: The label keys and values.
        :type labels: dict[str, Optional[str]]
        """
        if not labels:
            return
        labels = LabelUpdateRequest(labels=labels).dict()
        self._raw_ops.label(audit_id, labels)

    def upload(self, audit_id: str, file_path: str, **kwargs) -> None:
        """Upload an Audit report.

        Raises a 404 error if no entity exists with that key.

        :param audit_id: The Audit id.
        :type audit_id: str
        :param file_path: The path to the audit report data on the local file
            system. The report data must be packages as a ``.tar.gz`` archive.
        :type file_path: str
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        with open(file_path, "rb") as fin:
            request = build_audits_upload_request(
                audit_id=audit_id,
                file=fin,
                headers=_headers,
                params=_params,
            )
            request.url = self._raw_ops._client.format_url(request.url)

            _stream = False
            pipeline_response: PipelineResponse = (
                self._raw_ops._client._pipeline.run(  # pylint: disable=protected-access
                    request, stream=_stream, **kwargs
                )
            )

        response = pipeline_response.http_response

        if response.status_code not in [200, 422]:
            map_error(
                status_code=response.status_code, response=response, error_map=error_map
            )
            raise HttpResponseError(response=response)

        if response.status_code == 422:
            if response.content:
                deserialized = response.json()
            else:
                deserialized = None
            raise ValueError(f"Unprocessable entity: {deserialized}")


class AuditproceduresOperations:
    """Operations on :class:`~dyff.schema.platform.AuditProcedure` entities.

    .. note::

      Do not instantiate this class. Access it through the
      ``.auditprocedures`` attribute of :class:`~dyff.client.Client`.
    """

    def __init__(self, _raw_ops: AuditproceduresOperationsGenerated):
        self._raw_ops = _raw_ops

    def download(self, path: str) -> BytesIO:
        """Fetch the source code of an audit procedure.

        The data is returned as an in-memory ``.tar.gz`` archive.

        :param path: The relative path to the source code in the git repository.
        :type path: str
        :return: The audit procedure source code as an in-memory ``.tar.gz`` archive.
        """
        stream = self._raw_ops.download(path)
        blob = bytearray()
        for chunk in stream:
            blob.extend(chunk)  # type: ignore
        return BytesIO(blob)


class DatasetsOperations:
    """Operations on :class:`~dyff.schema.platform.Dataset` entities.

    .. note::

      Do not instantiate this class. Access it through the
      ``.datasets`` attribute of :class:`~dyff.client.Client`.
    """

    def __init__(self, _raw_ops: DatasetsOperationsGenerated):
        self._raw_ops = _raw_ops

    def get(self, dataset_id: str) -> Dataset:
        """Get a Dataset by its key.

        :param dataset_id: The dataset key
        :type dataset_id: str
        :return: The Dataset with the given key.
        """
        return Dataset.parse_obj(self._raw_ops.get(dataset_id))

    def delete(self, dataset_id: str) -> Status:
        """Mark a Dataset for deletion.

        :param dataset_id: The dataset key
        :type dataset_id: str
        :return: The resulting status of the entity
        :rtype: dyff.schema.platform.Status
        """
        return Status.parse_obj(self._raw_ops.delete(dataset_id))

    def query(
        self,
        *,
        id: Optional[str] = None,
        account: Optional[str] = None,
        status: Optional[str] = None,
        reason: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
        name: Optional[str] = None,
    ) -> list[Dataset]:
        """Get all Datasets matching a query. The query is a set of equality
        constraints specified as key-value pairs.

        :keyword id:
        :paramtype id: str
        :keyword account:
        :paramtype account: str
        :keyword status:
        :paramtype status: str
        :keyword reason:
        :paramtype reason: str
        :keyword labels:
        :paramtype labels: dict[str, str]
        :keyword name: Default value is None.
        :paramtype name: str
        :return: list of ``Dataset`` resources satisfying the query.
        :rtype: list[Dataset]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return [
            Dataset.parse_obj(obj)
            for obj in self._raw_ops.query(
                id=id,
                account=account,
                status=status,
                reason=reason,
                labels=_encode_labels(labels),
                name=name,
            )
        ]

    def label(self, dataset_id: str, labels: dict[str, Optional[str]]) -> None:
        """Label the specified Dataset with key-value pairs (stored in
        the ``.labels`` field of the resource).

        Providing ``None`` for the value deletes the label.

        See :class:`~dyff.schema.platform.Label` for a description of the
        constraints on label keys and values.

        :param dataset_id: The ID of the Dataset to label.
        :type dataset_id: str
        :param labels: The label keys and values.
        :type labels: dict[str, Optional[str]]
        """
        if not labels:
            return
        labels = LabelUpdateRequest(labels=labels).dict()
        self._raw_ops.label(dataset_id, labels)

    def strata(self, dataset: Dataset | str) -> pandas.DataFrame:
        """Fetch the strata of the dataset.

        :param dataset: The identifier of the dataset
        :type dataset: Dataset | str
        :return: The strata of the dataset
        """
        dataset = _require_id(dataset)
        stream = self._raw_ops.strata(dataset)
        blob = bytearray()
        for chunk in stream:
            blob.extend(chunk)  # type: ignore
        with BytesIO(blob) as fin:
            return pandas.read_parquet(fin)

    def data(self, dataset: Dataset | str) -> pyarrow.dataset.Scanner:
        """Fetch the raw data of the dataset.

        The data is returned as a read-once pyarrow Dataset.

        :param dataset: The identifier of the dataset
        :type dataset: Dataset | str
        :return: The strata of the dataset
        """
        dataset = _require_id(dataset)
        stream = self._raw_ops.data(dataset)
        buffer = bytearray()
        for chunk in stream:
            buffer.extend(chunk)  # type: ignore
        with pyarrow.ipc.open_stream(buffer) as reader:
            # Note: We should be able to pass 'reader' to 'from_batches()' directly
            # because it's already a batch generator, but if we do that then the
            # generator sometimes blocks program exit for some reason.
            def batch_generator():
                yield from reader

            return pyarrow.dataset.Scanner.from_batches(
                batch_generator(), schema=reader.schema
            )

    def create(self, dataset_request: DatasetCreateRequest) -> Dataset:
        """Create a Dataset.

        .. note::
            This operation may incur compute costs.

        :param dataset_request: The dataset request specification.
        :type dataset_request: DatasetCreateRequest
        :return: A full Dataset entity with .id and other properties set.
        """
        return Dataset.parse_obj(self._raw_ops.create(dataset_request.dict()))

    def create_arrow_dataset(
        self, dataset_directory: str, *, account: str, name: str
    ) -> Dataset:
        """Create a Dataset resource describing an existing Arrow dataset.

        Internally, constructs a ``DatasetCreateRequest`` using information
        obtained from the Arrow dataset, then calls ``create()`` with the
        constructed request.

        Typical usage::

            dataset = client.datasets.create_arrow_dataset(dataset_directory, ...)
            client.datasets.upload_arrow_dataset(dataset, dataset_directory)

        :param dataset_directory: The root directory of the Arrow dataset.
        :type dataset_directory: str
        :keyword account: The account that will own the Dataset resource.
        :type account: str
        :keyword name: The name of the Dataset resource.
        :type name: str
        :returns: The complete Dataset resource.
        :rtype: Dataset
        """
        dataset_path = pathlib.Path(dataset_directory)
        ds = arrow.open_dataset(str(dataset_path))
        file_paths = list(ds.files)
        artifact_paths = [
            str(pathlib.Path(file_path).relative_to(dataset_path))
            for file_path in file_paths
        ]
        artifacts = [
            Artifact(
                kind="parquet",
                path=artifact_path,
                digest=Digest(
                    md5=binary.encode(binary.file_digest("md5", file_path)),
                ),
            )
            for file_path, artifact_path in zip(file_paths, artifact_paths)
        ]
        schema = DataSchema(
            arrowSchema=arrow.encode_schema(ds.schema),
        )
        request = DatasetCreateRequest(
            account=account,
            name=name,
            artifacts=artifacts,
            schema=schema,
        )
        return self.create(request)

    def upload_arrow_dataset(self, dataset: Dataset, dataset_directory: str) -> None:
        """Uploads the data files in an existing Arrow dataset for which a
        Dataset resource has already been created.

        Typical usage::

            dataset = client.datasets.create_arrow_dataset(dataset_directory, ...)
            client.datasets.upload_arrow_dataset(dataset, dataset_directory)

        :param dataset: The Dataset resource for the Arrow dataset.
        :type dataset: Dataset
        :param dataset_directory: The root directory of the Arrow dataset.
        :type dataset_directory: str
        """
        if any(artifact.digest.md5 is None for artifact in dataset.artifacts):
            raise ValueError("artifact.digest.md5 must be set for all artifacts")
        for artifact in dataset.artifacts:
            assert artifact.digest.md5 is not None
            file_path = pathlib.Path(dataset_directory) / artifact.path
            put_url_json = self._raw_ops.upload(dataset.id, artifact.path)
            print(put_url_json)
            put_url = StorageSignedURL.parse_obj(put_url_json)
            if put_url.method != "PUT":
                raise ValueError(f"expected a PUT URL; got {put_url.method}")
            with open(file_path, "rb") as fin:
                headers = {
                    "content-md5": artifact.digest.md5,
                }
                headers.update(put_url.headers)
                response = httpx.put(put_url.url, data=fin, headers=headers)  # type: ignore
                print(response.request)
                print(response.request.headers)
                response.raise_for_status()
        self._raw_ops.finalize(dataset.id)


class EvaluationsOperations:
    """Operations on :class:`~dyff.schema.platform.Evaluation` entities.

    .. note::

      Do not instantiate this class. Access it through the
      ``.evaluations`` attribute of :class:`~dyff.client.Client`.
    """

    def __init__(self, _raw_ops: EvaluationsOperationsGenerated):
        self._raw_ops = _raw_ops

    def get(self, evaluation_id: str) -> Evaluation:
        """Get an Evaluation by its key.

        :param evaluation_id: The evaluation id
        :type evaluation_id: str
        :return: The Evaluation with the given key.
        """
        return Evaluation.parse_obj(self._raw_ops.get(evaluation_id))

    def delete(self, evaluation_id: str) -> Status:
        """Mark an Evaluation for deletion.

        :param evaluation_id: The evaluation key
        :type evaluation_id: str
        :return: The resulting status of the entity
        :rtype: dyff.schema.platform.Status
        """
        return Status.parse_obj(self._raw_ops.delete(evaluation_id))

    def query(
        self,
        *,
        id: Optional[str] = None,
        account: Optional[str] = None,
        status: Optional[str] = None,
        reason: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
        dataset: Optional[str] = None,
        inferenceService: Optional[str] = None,
        inferenceServiceName: Optional[str] = None,
        model: Optional[str] = None,
        modelName: Optional[str] = None,
    ) -> list[Evaluation]:
        """Get all Evaluations matching a query. The query is a set of equality
        constraints specified as key-value pairs.

        :keyword id:
        :paramtype id: str
        :keyword account:
        :paramtype account: str
        :keyword status:
        :paramtype status: str
        :keyword reason:
        :paramtype reason: str
        :keyword labels:
        :paramtype labels: dict[str, str]
        :keyword dataset:
        :paramtype dataset: str
        :keyword inferenceService:
        :paramtype inferenceService: str
        :keyword inferenceServiceName:
        :paramtype inferenceServiceName: str
        :keyword model:
        :paramtype model: str
        :keyword modelName:
        :paramtype modelName: str
        :return: list of ``Evaluation`` resources satisfying the query.
        :rtype: list[Evaluation]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return [
            Evaluation.parse_obj(obj)
            for obj in self._raw_ops.query(
                id=id,
                account=account,
                status=status,
                reason=reason,
                labels=_encode_labels(labels),
                dataset=dataset,
                inference_service=inferenceService,
                inference_service_name=inferenceServiceName,
                model=model,
                model_name=modelName,
            )
        ]

    def label(self, evaluation_id: str, labels: dict[str, Optional[str]]) -> None:
        """Label the specified Evaluation with key-value pairs (stored in
        the ``.labels`` field of the resource).

        Providing ``None`` for the value deletes the label.

        See :class:`~dyff.schema.platform.Label` for a description of the
        constraints on label keys and values.

        :param evaluation_id: The ID of the Evaluation to label.
        :type evaluation_id: str
        :param labels: The label keys and values.
        :type labels: dict[str, Optional[str]]
        """
        if not labels:
            return
        labels = LabelUpdateRequest(labels=labels).dict()
        self._raw_ops.label(evaluation_id, labels)

    def create(self, evaluation_request: EvaluationCreateRequest) -> Evaluation:
        """Create an Evaluation.

        .. note::
            This operation will incur compute costs.

        :param evaluation_request: The evaluation request specification.
        :type evaluation_request: EvaluationCreateRequest
        :return: A full Evaluation entity with .id and other properties set.
        """
        return Evaluation.parse_obj(self._raw_ops.create(evaluation_request.dict()))


class InferenceservicesOperations:
    """Operations on :class:`~dyff.schema.platform.InferenceService` entities.

    .. note::

      Do not instantiate this class. Access it through the
      ``.inferenceservices`` attribute of :class:`~dyff.client.Client`.
    """

    def __init__(self, _raw_ops: InferenceservicesOperationsGenerated):
        self._raw_ops = _raw_ops

    def get(self, service_id: str) -> InferenceService:
        """Get an InferenceService by its key.

        :param service_id: The inference service id
        :type service_id: str
        :return: The InferenceService with the given key.
        """
        return InferenceService.parse_obj(self._raw_ops.get(service_id))

    def delete(self, service_id: str) -> Status:
        """Mark an InferenceService for deletion.

        :param service_id: The inference service key
        :type service_id: str
        :return: The resulting status of the entity
        :rtype: dyff.schema.platform.Status
        """
        return Status.parse_obj(self._raw_ops.delete(service_id))

    def query(
        self,
        *,
        id: Optional[str] = None,
        account: Optional[str] = None,
        status: Optional[str] = None,
        reason: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
        name: Optional[str] = None,
        model: Optional[str] = None,
        modelName: Optional[str] = None,
    ) -> list[InferenceService]:
        """Get all InferenceServices matching a query. The query is a set of equality
        constraints specified as key-value pairs.

        :keyword id:
        :paramtype id: str
        :keyword account:
        :paramtype account: str
        :keyword status:
        :paramtype status: str
        :keyword reason:
        :paramtype reason: str
        :keyword labels:
        :paramtype labels: dict[str, str]
        :keyword name:
        :paramtype name: str
        :keyword model:
        :paramtype model: str
        :keyword modelName:
        :paramtype modelName: str
        :return: list of ``InferenceService`` resources satisfying the query.
        :rtype: list[InferenceService]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return [
            InferenceService.parse_obj(obj)
            for obj in self._raw_ops.query(
                id=id,
                account=account,
                status=status,
                reason=reason,
                labels=_encode_labels(labels),
                name=name,
                model=model,
                model_name=modelName,
            )
        ]

    def label(self, service_id: str, labels: dict[str, Optional[str]]) -> None:
        """Label the specified InferenceService with key-value pairs (stored in
        the ``.labels`` field of the resource).

        Providing ``None`` for the value deletes the label.

        See :class:`~dyff.schema.platform.Label` for a description of the
        constraints on label keys and values.

        :param service_id: The ID of the InferenceService to label.
        :type service_id: str
        :param labels: The label keys and values.
        :type labels: dict[str, Optional[str]]
        """
        if not labels:
            return
        labels = LabelUpdateRequest(labels=labels).dict()
        self._raw_ops.label(service_id, labels)

    def create(
        self, inference_service_request: InferenceServiceCreateRequest
    ) -> InferenceService:
        """Create an InferenceService.

        .. note::
            This operation may incur compute costs.

        :param inference_service_request: The inference service request specification.
        :type inference_service_request: InferenceServiceCreateRequest
        :return: A full InferenceService entity with .id and other properties set.
        """
        return InferenceService.parse_obj(
            self._raw_ops.create(inference_service_request.dict())
        )


class InferencesessionsOperations:
    """Operations on :class:`~dyff.schema.platform.Inferencesession` entities.

    .. note::

      Do not instantiate this class. Access it through the
      ``.inferencesessions`` attribute of :class:`~dyff.client.Client`.
    """

    def __init__(self, _raw_ops: InferencesessionsOperationsGenerated):
        self._raw_ops = _raw_ops

    def get(self, session_id: str) -> InferenceSession:
        """Get an InferenceSession by its key.

        :param session_id: The inference session id
        :type session_id: str
        :return: The InferenceSession with the given key.
        """
        return InferenceSession.parse_obj(self._raw_ops.get(session_id))

    def delete(self, session_id: str) -> Status:
        """Mark an InferenceSession for deletion.

        :param session_id: The inference session key
        :type session_id: str
        :return: The resulting status of the entity
        :rtype: dyff.schema.platform.Status
        """
        return Status.parse_obj(self._raw_ops.delete(session_id))

    def query(
        self,
        *,
        id: Optional[str] = None,
        account: Optional[str] = None,
        status: Optional[str] = None,
        reason: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
        name: Optional[str] = None,
        inferenceService: Optional[str] = None,
        inferenceServiceName: Optional[str] = None,
        model: Optional[str] = None,
        modelName: Optional[str] = None,
    ) -> list[InferenceSession]:
        """Get all InferenceSessions matching a query. The query is a set of equality
        constraints specified as key-value pairs.

        :keyword id:
        :paramtype id: str
        :keyword account:
        :paramtype account: str
        :keyword status:
        :paramtype status: str
        :keyword reason:
        :paramtype reason: str
        :keyword labels:
        :paramtype labels: dict[str, str]
        :keyword name:
        :paramtype name: str
        :keyword model:
        :paramtype model: str
        :keyword modelName:
        :paramtype modelName: str
        :return: list of ``InferenceSession`` resources satisfying the query.
        :rtype: list[InferenceSession]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return [
            InferenceSession.parse_obj(obj)
            for obj in self._raw_ops.query(
                id=id,
                account=account,
                status=status,
                reason=reason,
                labels=_encode_labels(labels),
                name=name,
                inference_service=inferenceService,
                inference_service_name=inferenceServiceName,
                model=model,
                model_name=modelName,
            )
        ]

    def label(self, session_id: str, labels: dict[str, Optional[str]]) -> None:
        """Label the specified InferenceSession with key-value pairs (stored in
        the ``.labels`` field of the resource).

        Providing ``None`` for the value deletes the label.

        See :class:`~dyff.schema.platform.Label` for a description of the
        constraints on label keys and values.

        :param session_id: The ID of the InferenceSession to label.
        :type session_id: str
        :param labels: The label keys and values.
        :type labels: dict[str, Optional[str]]
        """
        if not labels:
            return
        labels = LabelUpdateRequest(labels=labels).dict()
        self._raw_ops.label(session_id, labels)

    def create(
        self, inference_session_request: InferenceSessionCreateRequest
    ) -> InferenceSessionAndToken:
        """Create an InferenceSession.

        .. note::
            This operation will incur compute costs.

        :param account_id: The account that will own the evaluation.
        :type account_id: str
        :param inference_session_request: The inference service request specification.
        :type inference_session_request: InferenceSessionCreateRequest
        :return: A full InferenceSession entity with .id and other properties set.
        """
        return InferenceSessionAndToken.parse_obj(
            self._raw_ops.create(inference_session_request.dict())
        )

    def client(
        self,
        session_id: str,
        token: str,
        *,
        interface: Optional[InferenceInterface] = None,
        endpoint: Optional[str] = None,
        input_adapter: Optional[Adapter] = None,
        output_adapter: Optional[Adapter] = None,
    ) -> InferenceSessionClient:
        """Create an InferenceSessionClient that interacts with the given
        inference session. The token should be one returned either from
        ``Client.inferencesessions.create()`` or from
        ``Client.inferencesessions.token(session_id)``.

        The inference endpoint in the session must also be specified, either
        directly through the ``endpoint`` argument or by specifying an
        ``interface``. Specifying ``interface`` will also use the input and
        output adapters from the interface. You can also specify these
        separately in the ``input_adapter`` and ``output_adapter``. The
        non-``interface`` arguments override the corresponding values in
        ``interface`` if both are specified.

        :param session_id: The inference session to connect to
        :type session_id: str
        :param token: An access token with permission to run inference against
            the session.
        :type token: str
        :param interface: The interface to the session. Either ``interface``
            or ``endpoint`` must be specified.
        :type interface: Optional[InferenceInterface]
        :param endpoint: The inference endpoint in the session to call. Either
            ``endpoint`` or ``interface`` must be specified.
        :type endpoint: str
        :param input_adapter: Optional input adapter, applied to the input
            before sending it to the session. Will override the input adapter
            from ``interface`` if both are specified.
        :type input_adapter: Optional[Adapter]
        :param output_adapter: Optional output adapter, applied to the output
            of the session before returning to the client. Will override the
            output adapter from ``interface`` if both are specified.
        :type output_adapter: Optional[Adapter]
        :return: An ``InferenceSessionClient`` that makes inference calls to
            the specified session.
        """
        if interface is not None:
            endpoint = endpoint or interface.endpoint
            if input_adapter is None:
                if interface.inputPipeline is not None:
                    input_adapter = create_pipeline(interface.inputPipeline)
            if output_adapter is None:
                if interface.outputPipeline is not None:
                    output_adapter = create_pipeline(interface.outputPipeline)
        if endpoint is None:
            raise ValueError("either 'endpoint' or 'interface' is required")
        return InferenceSessionClient(
            session_id=session_id,
            token=token,
            dyff_api_endpoint=self._raw_ops._client._base_url,
            inference_endpoint=endpoint,
            input_adapter=input_adapter,
            output_adapter=output_adapter,
        )

    def ready(self, session_id: str) -> bool:
        """Return True if the session is ready to receive inference input.

        The readiness probe is expected to fail with status codes 404 or 503,
        as these will occur at times during normal session start-up. The
        ``ready()`` method returns False in these cases. Any other status
        codes will raise an ``HttpResponseError``.

        :param str session_id: The ID of the session.
        :raises HttpResponseError:
        """
        try:
            self._raw_ops.ready(session_id)
        except HttpResponseError as ex:
            if ex.status_code in [404, 503]:
                return False
            else:
                raise
        return True

    def terminate(self, session_id: str) -> Status:
        """Terminate a session.

        :param session_id: The inference session key
        :type session_id: str
        :return: The resulting status of the entity
        :rtype: dyff.schema.platform.Status
        :raises HttpResponseError:
        """
        return Status.parse_obj(self._raw_ops.terminate(session_id))

    def token(self, session_id: str) -> str:
        """Create a session token.

        The session token is a short-lived token that allows the bearer to
        make inferences with the session (via an ``InferenceSessionClient``)
        and to call ``ready()``, ``get()``, and ``terminate()`` on the session.

        :param str session_id: The ID of the session.
        :raises HttpResponseError:
        """
        return str(self._raw_ops.token(session_id))


class ModelsOperations:
    """Operations on :class:`~dyff.schema.platform.Model` entities.

    .. note::

      Do not instantiate this class. Access it through the
      ``.models`` attribute of :class:`~dyff.client.Client`.
    """

    def __init__(self, _raw_ops: ModelsOperationsGenerated):
        self._raw_ops = _raw_ops

    def get(self, model_id: str) -> Model:
        """Get a Model by its key.

        :param model_id: The inference service id
        :type model_id: str
        :return: The Model with the given key.
        """
        return Model.parse_obj(self._raw_ops.get(model_id))

    def delete(self, model_id: str) -> Status:
        """Mark a Model for deletion.

        :param model_id: The model key
        :type model_id: str
        :return: The resulting status of the entity
        :rtype: dyff.schema.platform.Status
        """
        return Status.parse_obj(self._raw_ops.delete(model_id))

    def query(
        self,
        *,
        id: Optional[str] = None,
        account: Optional[str] = None,
        status: Optional[str] = None,
        reason: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
        name: Optional[str] = None,
    ) -> list[Model]:
        """Get all Models matching a query. The query is a set of equality
        constraints specified as key-value pairs.

        :keyword id:
        :paramtype id: str
        :keyword account:
        :paramtype account: str
        :keyword status:
        :paramtype status: str
        :keyword reason:
        :paramtype reason: str
        :keyword labels:
        :paramtype labels: dict[str, str]
        :keyword name:
        :paramtype name: str
        :return: list of ``Model`` resources satisfying the query.
        :rtype: list[Model]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return [
            Model.parse_obj(obj)
            for obj in self._raw_ops.query(
                id=id,
                account=account,
                status=status,
                reason=reason,
                labels=_encode_labels(labels),
                name=name,
            )
        ]

    def label(self, model_id: str, labels: dict[str, Optional[str]]) -> None:
        """Label the specified Model with key-value pairs (stored in
        the ``.labels`` field of the resource).

        Providing ``None`` for the value deletes the label.

        See :class:`~dyff.schema.platform.Label` for a description of the
        constraints on label keys and values.

        :param model_id: The ID of the Model to label.
        :type model_id: str
        :param labels: The label keys and values.
        :type labels: dict[str, Optional[str]]
        """
        if not labels:
            return
        labels = LabelUpdateRequest(labels=labels).dict()
        self._raw_ops.label(model_id, labels)

    def create(self, model_request: ModelCreateRequest) -> Model:
        """Create a Model.

        .. note::
            This operation will incur compute costs.

        :param model_request: The model specification.
        :type model: Model
        :return: A full Model entity with .id and other properties set.
        """
        return Model.parse_obj(self._raw_ops.create(model_request.dict()))


class ReportsOperations:
    """Operations on :class:`~dyff.schema.platform.Report` entities.

    .. note::

      Do not instantiate this class. Access it through the
      ``.reports`` attribute of :class:`~dyff.client.Client`.
    """

    def __init__(self, _raw_ops: ReportsOperationsGenerated):
        self._raw_ops = _raw_ops

    def get(self, report_id: str) -> Report:
        """Get a Report by its key.

        :param report_id: The report id
        :type report_id: str
        :return: The Report with the given key.
        """
        return Report.parse_obj(self._raw_ops.get(report_id))

    def delete(self, report_id: str) -> Status:
        """Mark a Report for deletion.

        :param report_id: The report key
        :type report_id: str
        :return: The resulting status of the entity
        :rtype: dyff.schema.platform.Status
        """
        return Status.parse_obj(self._raw_ops.delete(report_id))

    def query(
        self,
        *,
        id: Optional[str] = None,
        account: Optional[str] = None,
        status: Optional[str] = None,
        reason: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
        report: Optional[str] = None,
        dataset: Optional[str] = None,
        evaluation: Optional[str] = None,
        inferenceService: Optional[str] = None,
        model: Optional[str] = None,
    ) -> list[Report]:
        """Get all Reports matching a query. The query is a set of equality
        constraints specified as key-value pairs.

        :keyword id:
        :paramtype id: str
        :keyword account:
        :paramtype account: str
        :keyword status:
        :paramtype status: str
        :keyword reason:
        :paramtype reason: str
        :keyword labels:
        :paramtype labels: dict[str, str]
        :keyword report:
        :paramtype report: str
        :keyword dataset:
        :paramtype dataset: str
        :keyword evaluation:
        :paramtype evaluation: str
        :keyword inferenceService:
        :paramtype inferenceService: str
        :keyword model:
        :paramtype model: str
        :return: list of ``Report`` resources satisfying the query.
        :rtype: list[Report]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        return [
            Report.parse_obj(obj)
            for obj in self._raw_ops.query(
                id=id,
                account=account,
                status=status,
                reason=reason,
                labels=_encode_labels(labels),
                report=report,
                dataset=dataset,
                evaluation=evaluation,
                inference_service=inferenceService,
                model=model,
            )
        ]

    def label(self, report_id: str, labels: dict[str, Optional[str]]) -> None:
        """Label the specified Report with key-value pairs (stored in
        the ``.labels`` field of the resource).

        Providing ``None`` for the value deletes the label.

        See :class:`~dyff.schema.platform.Label` for a description of the
        constraints on label keys and values.

        :param report_id: The ID of the Report to label.
        :type report_id: str
        :param labels: The label keys and values.
        :type labels: dict[str, Optional[str]]
        """
        if not labels:
            return
        labels = LabelUpdateRequest(labels=labels).dict()
        self._raw_ops.label(report_id, labels)

    def create(self, report_request: ReportCreateRequest) -> Report:
        """Create a Report.

        .. note::
            This operation will incur compute costs.

        :param report_request: The report specification.
        :type report: ReportCreateRequest
        :return: A full Report entity with .id and other properties set.
        """
        return Report.parse_obj(self._raw_ops.create(report_request.dict()))

    def data(self, report: Report | str) -> pandas.DataFrame:
        """Fetch the output data of the report.

        :param report: The identifier of the report
        :type report: Report | str
        :return: The output data of the report
        """
        report = _require_id(report)
        stream = self._raw_ops.data(report)
        blob = bytearray()
        for chunk in stream:
            blob.extend(chunk)  # type: ignore
        with BytesIO(blob) as fin:
            return pandas.read_parquet(fin)


class _APIKeyCredential(TokenCredential):
    def __init__(self, *, api_key: str):
        self.api_key = api_key

    def get_token(
        self,
        *scopes: str,
        claims: Optional[str] = None,
        tenant_id: Optional[str] = None,
        **kwargs: Any,
    ) -> AccessToken:
        return AccessToken(self.api_key, -1)


class Client:
    """The Python client for the Dyff Platform API.

    API operations are grouped by the resource type that they manipulate. For
    example, all operations on ``Evaluation`` resources are accessed like
    ``client.evaluations.create()``.

    The Python API functions may have somewhat different behavior from the
    corresponding API endpoints, and the Python client also adds several
    higher-level API functions that are implemented with multiple endpoint
    calls.
    """

    def __init__(
        self,
        *,
        api_key: str,
        endpoint: Optional[str] = None,
        verify_ssl_certificates: bool = True,
    ):
        """
        :param str api_key: An API token to use for authentication.
        :param str endpoint: The URL where the Dyff Platform API is hosted.
            Defaults to the UL DSRI-hosted Dyff instance.
        :param bool verify_ssl_certificates: You can disable certificate
            verification for testing; you should do this only if you have
            also changed ``endpoint`` to point to a trusted local server.
        """

        if endpoint is None:
            endpoint = "https://apis.alignmentlabs.ai/dyff/v0"
        credential = _APIKeyCredential(api_key=api_key)
        authentication_policy = BearerTokenCredentialPolicy(credential)
        self._raw = RawClient(
            endpoint=endpoint,
            credential=credential,
            authentication_policy=authentication_policy,
        )

        # We want the ability to disable SSL certificate verification for testing
        # on localhost. It should be possible to do this via the Configuration object:
        # e.g., config.<some_field> = azure.core.configuration.ConnectionConfiguration(connection_verify=False)
        #
        # The docs state that the ConnectionConfiguration class is "Found in the Configuration object."
        # https://learn.microsoft.com/en-us/python/api/azure-core/azure.core.configuration.connectionconfiguration?view=azure-python
        #
        # But at no point do they say what the name of the field should be! The
        # docs for azure.core.configuration.Configuration don't mention any
        # connection configuration. The field is called 'connection_config' in the
        # _transport member of _pipeline, but _transport will not pick up the
        # altered ConnectionConfiguration if it is set on 'config.connection_config'
        #
        # Example:
        # client._config.connection_config = ConnectionConfiguration(connection_verify=False)
        # [in Client:]
        # >>> print(self._config.connection_config.verify)
        # False
        # >> print(self._pipeline._transport.connection_config.verify)
        # True
        self._raw._client._pipeline._transport.connection_config.verify = (  # type: ignore
            verify_ssl_certificates
        )

        self._audits = AuditsOperations(self._raw.audits)
        self._auditprocedures = AuditproceduresOperations(self._raw.auditprocedures)
        self._datasets = DatasetsOperations(self._raw.datasets)
        self._evaluations = EvaluationsOperations(self._raw.evaluations)
        self._inferenceservices = InferenceservicesOperations(
            self._raw.inferenceservices
        )
        self._inferencesessions = InferencesessionsOperations(
            self._raw.inferencesessions
        )
        self._models = ModelsOperations(self._raw.models)
        self._reports = ReportsOperations(self._raw.reports)

    @property
    def raw(self) -> RawClient:
        """The "raw" API client, which can be used to send JSON requests directly."""
        return self._raw

    @property
    def audits(self) -> AuditsOperations:
        """Operations on :class:`~dyff.schema.platform.Audit` entities."""
        return self._audits

    @property
    def auditprocedures(self) -> AuditproceduresOperations:
        """Operations on :class:`~dyff.schema.platform.AuditProcedure` entities."""
        return self._auditprocedures

    @property
    def datasets(self) -> DatasetsOperations:
        """Operations on :class:`~dyff.schema.platform.Dataset` entities."""
        return self._datasets

    @property
    def evaluations(self) -> EvaluationsOperations:
        """Operations on :class:`~dyff.schema.platform.Evaluation` entities."""
        return self._evaluations

    @property
    def inferenceservices(self) -> InferenceservicesOperations:
        """Operations on :class:`~dyff.schema.platform.InferenceService` entities."""
        return self._inferenceservices

    @property
    def inferencesessions(self) -> InferencesessionsOperations:
        """Operations on :class:`~dyff.schema.platform.InferenceSession` entities."""
        return self._inferencesessions

    @property
    def models(self) -> ModelsOperations:
        """Operations on :class:`~dyff.schema.platform.Model` entities."""
        return self._models

    @property
    def reports(self) -> ReportsOperations:
        """Operations on :class:`~dyff.schema.platform.Report` entities."""
        return self._reports


__all__ = [
    "Client",
    "InferenceSessionClient",
    "RawClient",
    "HttpResponseError",
]



#!/usr/bin/env python3
# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import base64
import os
import secrets

import kubernetes.client
import kubernetes.config

# https://stackoverflow.com/a/51035715
if os.environ.get("KUBERNETES_SERVICE_HOST"):
    kubernetes.config.load_incluster_config()
else:
    kubernetes.config.load_kube_config()

api = kubernetes.client.CoreV1Api()
# token_urlsafe() raises 'binascii.Error: Incorrect padding' if you later try
# to base64 decode it
token = base64.urlsafe_b64encode(secrets.token_bytes(32)).decode()
body = {"stringData": {"ALIGNMENTLABS_API_KEY_SIGNING_SECRET": token}}
api.patch_namespaced_secret(name="api-key-signing", namespace="default", body=body)



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import re
from typing import Dict, List, Union, api

import datasets
import gcsfs
import pyarrow
from absl import logging

from ...core.config import config
from .. import storage
from . import arrowtypes, io
from . import text as text_data

DatasetType = Union[
    datasets.Dataset,
    datasets.IterableDataset,
    datasets.DatasetDict,
    datasets.IterableDatasetDict,
]


def fetch(spec: api.DataSource):
    """Fetch a Huggingface dataset into storage."""
    if spec.sourceKind != api.DataSources.huggingface:
        raise ValueError(f"requires spec.sourceKind = {api.DataSources.huggingface}")
    prefix = "https://huggingface.co/datasets/"
    if spec.source.startswith(prefix):
        dataset = spec.source[len(prefix) :]
    else:
        raise ValueError(
            "expected spec.source like 'https://huggingface.co/datasets/<dataset>'"
        )
    fs = gcsfs.GCSFileSystem()
    output_dir = storage.paths.datasource_root(spec.id)
    builder = datasets.load_dataset_builder(dataset)
    builder.download_and_prepare(
        output_dir, storage_options=fs.storage_options, file_format="parquet"
    )


def get_dataset_info(
    datasource_id: api.DataSource | api.EntityKey | str,
) -> datasets.DatasetInfo:
    if isinstance(datasource_id, api.DataSource):
        datasource_id = datasource_id.key
    elif not isinstance(datasource_id, api.EntityKey):
        datasource_id = api.EntityKey.from_str(datasource_id)
    path = storage.paths.datasource_root(datasource_id)
    fs = gcsfs.GCSFileSystem()
    return datasets.DatasetInfo.from_directory(path, storage_options=fs.storage_options)


class TextSpans:
    """Text data converter that augments tokens and token-level tags with
    detokenized text and span-level tags.
    """

    def __init__(self, *, tag_fields: List[str], span_fields: Dict[str, str]):
        """
        Parameters:
          tag_fields: The token-level tag fields to preserve
          span_fields: The token-level tag fields for which to generate
            corresponding span-level tags. The key is the token-level tag field
            name and the value is the name to assign to the generated span-level
            tag field.
        """
        self.tag_fields = tag_fields
        self.span_fields = span_fields
        self._all_fields = set(self.tag_fields).union(self.span_fields)
        all_field_names = set(self.tag_fields).union(self.span_fields.values())
        if len(all_field_names) != (len(self.tag_fields) + len(self.span_fields)):
            raise ValueError("duplicate field names in result schema")
        schema = [
            pyarrow.field("text", pyarrow.string()),
            pyarrow.field("tokens", pyarrow.list_(pyarrow.string())),
        ]
        for f in self.tag_fields:
            schema.append(pyarrow.field(f, pyarrow.list_(pyarrow.string())))
        for v in self.span_fields.values():
            schema.append(pyarrow.field(v, pyarrow.list_(arrowtypes.TextSpan())))
        self._schema = pyarrow.schema(schema)

    @property
    def num_instance_features(self) -> int:
        """The number of features in a single instance."""
        return len(self._schema.names)

    @property
    def partition_schema(self) -> pyarrow.Schema:
        """The ``pyarrow.schema`` describing the subset of the features that
        should be represented using the ``pyarrow.dataset`` "partitioning"
        mechanism.
        """
        return None

    @property
    def schema(self) -> pyarrow.Schema:
        """The ``pyarrow.schema`` describing the full set of features."""
        return self._schema

    def __call__(self, dataset_info, example):
        tag_dicts = {
            k: dataset_info.features[k].feature._int2str for k in self._all_fields
        }
        result = {}
        text = text_data.detokenize(example["tokens"])
        result["text"] = text
        result["tokens"] = list(example["tokens"])
        for f in self.tag_fields:
            result[f] = [tag_dicts[f][t] for t in example[f]]
        for k, v in self.span_fields.items():
            result[v] = [
                span.to_dict()
                for span in text_data.compute_spans(
                    text, example["tokens"], [tag_dicts[k][t] for t in example[k]]
                )
            ]
        logging.debug(result)
        return result


class Ingester:
    def __init__(self, dataset: api.Dataset | api.EntityKey | str, converter):
        if isinstance(dataset, api.Dataset):
            self.dataset = dataset
        else:
            self.dataset = api.get_dataset(dataset)
        self.converter = converter

        if self.dataset.sourceKind != api.DataSources.huggingface:
            raise ValueError(
                f"requires dataset.sourceKind = {api.DataSources.huggingface}"
            )

        self.data_source = api.get_data_source(self.dataset.source)
        self.dataset_info = get_dataset_info(self.data_source)

        self._partitions = []
        self._schema = [pyarrow.field("_index_", pyarrow.int64())]

        if self.dataset_info.splits:
            self._schema.append(pyarrow.field("split", pyarrow.string()))
            self._partitions.append(pyarrow.field("split", pyarrow.string()))

        if self.converter.partition_schema is not None:
            self._partitions.extend(
                pyarrow.field(name, type_)
                for name, type_ in zip(
                    self.converter.partition_schema.names,
                    self.converter.partition_schema.types,
                )
            )
        if self.converter.schema is not None:
            self._schema.extend(
                pyarrow.field(name, type_)
                for name, type_ in zip(
                    self.converter.schema.names, self.converter.schema.types
                )
            )

    @property
    def num_instance_features(self) -> int:
        """The number of features in a single instance."""
        return len(self._schema)

    @property
    def partition_schema(self) -> pyarrow.Schema:
        """The ``pyarrow.schema`` describing the subset of the features that
        should be represented using the ``pyarrow.dataset`` "partitioning"
        mechanism.
        """
        return None if len(self._partitions) == 0 else pyarrow.schema(self._partitions)

    @property
    def schema(self) -> pyarrow.Schema:
        """The ``pyarrow.schema`` describing the full set of features."""
        return pyarrow.schema(self._schema)

    def __iter__(self):
        data_source_path = storage.paths.datasource_root(self.data_source.key)
        logging.debug(f"data_source_path: {data_source_path}")
        blobs = [blob.name for blob in storage.gcs_list_dir(data_source_path)]
        logging.debug(f"blobs: {blobs}")

        next_index = 0
        if self.dataset_info.splits:
            for split in self.dataset_info.splits.values():
                logging.info(split.name)
                pattern = f"/{split.dataset_name}-{split.name}.*\\.parquet$"
                split_blobs = [blob for blob in blobs if re.search(pattern, blob)]
                logging.info(split_blobs)
                ds = io.open_dataset(
                    split_blobs, filesystem=config.resources.datasources.storage.url
                )

                for batch in ds.to_batches():
                    for item in batch.to_pylist():
                        converted = self.converter(self.dataset_info, item)
                        converted["_index_"] = next_index
                        converted["split"] = split.name
                        next_index += 1
                        logging.debug(converted)
                        yield converted
        else:
            logging.info("<no splits>")
            pattern = f"/.*\\.parquet$"
            blobs = [blob for blob in blobs if re.search(pattern, blob)]
            logging.info(blobs)
            ds = io.open_dataset(
                blobs, filesystem=config.resources.datasources.storage.url
            )

            for batch in ds.to_batches():
                for item in batch.to_pylist():
                    converted = self.converter(self.dataset_info, item)
                    converted["_index_"] = next_index
                    next_index += 1
                    logging.debug(converted)
                    yield converted



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from typing import Union


def sanitize_relative_file_path(path: Union[Path, str]) -> Path:
    """Raises a ``ValueError`` if the provided path is not a relative path that
    is safe to extract files into. An unsafe path is one that either:
      1. is an absolute path
      2. references paths outside of its "root" directory (e.g., "foo/../../bar")

    Returns:
      The fully-resolved relative path
    """
    path = Path(path)
    if path.is_absolute():
        raise ValueError("not a relative path")
    cwd = Path(".")
    # This will raise a ValueError if 'path' refers to a directory that is
    # not a subdirectory of 'cwd' (which, since 'path' is a relative path, means
    # that 'somedir/path' would be outside of 'somedir', which is an exploit).
    return path.resolve().relative_to(cwd.resolve())



# Copyright UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import os

import fastapi
import jinja2
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles

from dyff.client import Client

DYFF_API_KEY = os.environ.get("DYFF_API_KEY")
dyff_api = Client(api_key=DYFF_API_KEY)


def _generate_unique_id(route: fastapi.routing.APIRoute) -> str:
    """Strip everything but the function name from the ``operationId`` fields."""
    return route.name


def check_route_permissions(request: fastapi.Request) -> None:
    pass


jenv = jinja2.Environment(
    loader=jinja2.PackageLoader("dyff.web"), autoescape=jinja2.select_autoescape()
)


app = fastapi.FastAPI(
    title="Dyff Cloud Console",
    generate_unique_id_function=_generate_unique_id,
    dependencies=[fastapi.Depends(check_route_permissions)],
)


this_dir = os.path.dirname(__file__)
static_dir = os.path.join(this_dir, "static")
app.mount("/static", StaticFiles(directory=static_dir), name="static")


@app.get("/")
def homepage() -> HTMLResponse:
    template = jenv.get_template("home.jinja")
    return HTMLResponse(template.render())


@app.get("/models/")
def models() -> HTMLResponse:
    template = jenv.get_template("models.jinja")
    return HTMLResponse(template.render())


@app.get("/models/data")
def models_all(model_id: str = None, account: str = "public") -> HTMLResponse:
    id = model_id if model_id != "" else None
    acc = account if account != "" else None
    models = dyff_api.models.query(id=id, account=acc)

    # Loop and generate random status to test dynamic status

    # Extract data in here for typing - no typing in .jinja
    headers = ["id", "name", "account", "source", "creationTime", "status"]
    body = [
        [m.id, m.name, m.account, m.source.kind, m.creationTime, m.status]
        for m in models
    ]

    template = jenv.get_template("table.jinja")
    return HTMLResponse(template.render(headers=headers, body=body, label="Models"))


@app.get("/reports/")
def reports() -> HTMLResponse:
    template = jenv.get_template("reports.jinja")
    return HTMLResponse(template.render())


@app.get("/reports/data")
def reports_all(report_id: str = None, account: str = None) -> HTMLResponse:
    id = report_id if report_id != "" else None
    acc = account if account != "" else None
    reports = dyff_api.reports.query(account=acc, id=id)

    # Extract data in here for typing - no typing in .jinja
    headers = ["id", "account", "evaluation", "creationTime", "status"]
    body = [[r.id, r.account, r.evaluation, r.creationTime, r.status] for r in reports]

    template = jenv.get_template("table.jinja")
    return HTMLResponse(template.render(headers=headers, body=body, label="Reports"))
